#!/usr/bin/env node
(() => {
    const defines = {};
    const entry = [null];
    function define(name, dependencies, factory) {
        defines[name] = { dependencies, factory };
        entry[0] = name;
    }
    define("require", ["exports"], (exports) => {
        Object.defineProperty(exports, "__cjsModule", { value: true });
        Object.defineProperty(exports, "default", { value: (name) => resolve(name) });
    });
    define("engine/animation/animation", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const linear = (s1, s2, amount) => s1 + ((s2 - s1) * amount);
        class Animation {
            constructor() {
                this.channels = new Map();
            }
            assert_key(key) {
                if (!this.channels.has(key)) {
                    this.channels.set(key, []);
                }
            }
            length(key) {
                this.assert_key(key);
                const current = this.channels.get(key);
                if (current.length === 0) {
                    return 0;
                }
                const last = current[current.length - 1];
                return last.timestamp;
            }
            add(key, frame) {
                this.assert_key(key);
                const current = this.channels.get(key);
                const deduped = current.filter(track => track.timestamp !== frame.timestamp);
                const next = [...deduped, frame].sort((a, b) => (a.timestamp < b.timestamp) ? -1 : 1);
                this.channels.set(key, next);
            }
            state(key, timestamp) {
                this.assert_key(key);
                const intersect = this.intersect(this.channels.get(key), timestamp);
                if (intersect) {
                    return this.linear(intersect, intersect.amount);
                }
                else {
                    return {};
                }
            }
            linear(intersect, amount) {
                const properties = {};
                for (const key of Object.keys(intersect.frame1.properties)) {
                    const frame1 = intersect.frame1.properties[key];
                    const frame2 = intersect.frame2.properties[key];
                    properties[key] = frame1.map((_, index) => linear(frame1[index], frame2[index], amount));
                }
                return properties;
            }
            intersect(frames, timestamp) {
                if (frames.length === 0) {
                    throw Error('Cannot intersect as no frames exist');
                }
                if (frames.length === 1) {
                    return {
                        amount: 0.0,
                        frame0: frames[0],
                        frame1: frames[0],
                        frame2: frames[0],
                        frame3: frames[0],
                    };
                }
                if (timestamp < frames[0].timestamp) {
                    return {
                        amount: 0.0,
                        frame0: frames[0],
                        frame1: frames[0],
                        frame2: (frames.length > 1) ? frames[1] : frames[0],
                        frame3: (frames.length > 2) ? frames[2] : (frames.length > 1) ? frames[1] : frames[0],
                    };
                }
                for (let i = 0; i < frames.length - 1; i++) {
                    const current = frames[i + 0];
                    const next = frames[i + 1];
                    if (timestamp >= current.timestamp && timestamp < next.timestamp) {
                        return {
                            amount: 1.0 - ((timestamp - next.timestamp) / (current.timestamp - next.timestamp)),
                            frame0: (i === 0) ? frames[i + 0] : frames[i - 1],
                            frame1: current,
                            frame2: next,
                            frame3: ((i + 1) === frames.length - 1) ? frames[i + 1] : frames[i + 2],
                        };
                    }
                }
                const last = frames.length - 1;
                return {
                    amount: 1.0,
                    frame0: (last === 0) ? frames[last] : frames[last - 1],
                    frame1: frames[last],
                    frame2: frames[last],
                    frame3: frames[last],
                };
            }
        }
        exports.Animation = Animation;
    });
    define("engine/animation/index", ["require", "exports", "engine/animation/animation"], function (require, exports, animation_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Animation = animation_1.Animation;
    });
    define("engine/math/angle", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Angle {
            static fromRadian(radian) {
                return (radian * 57.29578);
            }
        }
        exports.Angle = Angle;
    });
    define("engine/math/radian", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Radian {
            static fromAngle(angle) {
                return (angle * 0.01745329);
            }
        }
        exports.Radian = Radian;
    });
    define("engine/math/vector3", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const f32 = { min: Number.MIN_VALUE, max: Number.MAX_VALUE };
        class Vector3 {
            constructor(x, y, z) {
                this.v = new Float32Array(3);
                this.v[0] = x;
                this.v[1] = y;
                this.v[2] = z;
            }
            toString() {
                return `[${this.v[0]}, ${this.v[1]}, ${this.v[2]}]`;
            }
            kind() {
                return 'Vector3';
            }
            clone() {
                return Vector3.clone(this);
            }
            get x() {
                return this.v[0];
            }
            get y() {
                return this.v[1];
            }
            get z() {
                return this.v[2];
            }
            set x(value) {
                this.v[0] = value;
            }
            set y(value) {
                this.v[1] = value;
            }
            set z(value) {
                this.v[2] = value;
            }
            length() {
                return Vector3.getLength(this);
            }
            lengthSq() {
                return Vector3.getLengthSq(this);
            }
            normalize() {
                return Vector3.normalize(this);
            }
            dot(v0) {
                return Vector3.dot(this, v0);
            }
            cross(v0) {
                return Vector3.cross(this, v0);
            }
            add(v0) {
                return Vector3.add(this, v0);
            }
            sub(v0) {
                return Vector3.sub(this, v0);
            }
            mul(v0) {
                return Vector3.mul(this, v0);
            }
            div(v0) {
                return Vector3.div(this, v0);
            }
            scale(s0) {
                return Vector3.scale(this, s0);
            }
            negate() {
                return Vector3.negate(this);
            }
            static zero() {
                return new Vector3(0.0, 0.0, 0.0);
            }
            static one() {
                return new Vector3(1.0, 1.0, 1.0);
            }
            static unitX() {
                return new Vector3(1.0, 0.0, 0.0);
            }
            static unitY() {
                return new Vector3(0.0, 1.0, 0.0);
            }
            static unitZ() {
                return new Vector3(0.0, 0.0, 1.0);
            }
            static left() {
                return new Vector3(-1.0, 0.0, 0.0);
            }
            static right() {
                return new Vector3(1.0, 0.0, 0.0);
            }
            static up() {
                return new Vector3(0.0, 1.0, 0.0);
            }
            static down() {
                return new Vector3(0.0, -1.0, 0.0);
            }
            static forward() {
                return new Vector3(0.0, 0.0, 1.0);
            }
            static backward() {
                return new Vector3(0.0, 0.0, -1.0);
            }
            static equals(v0, v1) {
                return (v0.v[0] === v1.v[0] &&
                    v0.v[1] === v1.v[1] &&
                    v0.v[2] === v1.v[2]);
            }
            static getLength(v0) {
                return Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]));
            }
            static getLengthSq(v0) {
                return ((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]));
            }
            static distance(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                const z = v0.v[2] - v1.v[2];
                return Math.sqrt((x * x) + (y * y) + (z * z));
            }
            static distanceSq(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                const z = v0.v[2] - v1.v[2];
                return ((x * x) + (y * y) + (z * z));
            }
            static dot(v0, v1) {
                return ((v0.v[0] * v1.v[0]) +
                    (v0.v[1] * v1.v[1]) +
                    (v0.v[2] * v1.v[2]));
            }
            static normalize(v0) {
                const len = 1.0 / Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]));
                return new Vector3(v0.v[0] * len, v0.v[1] * len, v0.v[2] * len);
            }
            static cross(v0, v1) {
                return new Vector3((v0.v[1] * v1.v[2]) - (v0.v[2] * v1.v[1]), (v0.v[2] * v1.v[0]) - (v0.v[0] * v1.v[2]), (v0.v[0] * v1.v[1]) - (v0.v[1] * v1.v[0]));
            }
            static reflect(v0, n0) {
                const dot = ((v0.v[0] * n0.v[0]) +
                    (v0.v[1] * n0.v[1]) +
                    (v0.v[2] * n0.v[2]));
                return new Vector3(v0.v[0] - ((2.0 * dot) * n0.v[0]), v0.v[1] - ((2.0 * dot) * n0.v[1]), v0.v[2] - ((2.0 * dot) * n0.v[2]));
            }
            static abs(v0) {
                return new Vector3(Math.abs(v0.v[0]), Math.abs(v0.v[1]), Math.abs(v0.v[2]));
            }
            static min(v0, v1) {
                return new Vector3((v0.v[0] < v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] < v1.v[1]) ? v0.v[1] : v1.v[1], (v0.v[2] < v1.v[2]) ? v0.v[2] : v1.v[2]);
            }
            static max(v0, v1) {
                return new Vector3((v0.v[0] > v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] > v1.v[1]) ? v0.v[1] : v1.v[1], (v0.v[2] > v1.v[2]) ? v0.v[2] : v1.v[2]);
            }
            static clamp(v0, min, max) {
                let x = v0.v[0];
                let y = v0.v[1];
                let z = v0.v[2];
                x = (x > max.v[0]) ? max.v[0] : x;
                x = (x < min.v[0]) ? min.v[0] : x;
                y = (y > max.v[1]) ? max.v[1] : y;
                y = (y < min.v[1]) ? min.v[1] : y;
                z = (z > max.v[2]) ? max.v[2] : z;
                z = (z < min.v[2]) ? min.v[2] : z;
                return new Vector3(x, y, z);
            }
            static lerp(v0, v1, amount) {
                return new Vector3(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount), v0.v[2] + ((v1.v[2] - v0.v[2]) * amount));
            }
            static barycentric(v0, v1, v2, amount0, amount1) {
                return new Vector3((v0.v[0] + (amount0 * (v1.v[0] - v0.v[0]))) + (amount1 * (v2.v[0] - v0.v[0])), (v0.v[1] + (amount0 * (v1.v[1] - v0.v[1]))) + (amount1 * (v2.v[1] - v0.v[1])), (v0.v[2] + (amount0 * (v1.v[2] - v0.v[2]))) + (amount1 * (v2.v[2] - v0.v[2])));
            }
            static smoothstep(v0, v1, amount) {
                amount = (amount > 1.0) ? 1.0 : ((amount < 0.0) ? 0.0 : amount);
                amount = (amount * amount) * (3.0 - (2.0 * amount));
                return new Vector3(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount), v0.v[2] + ((v1.v[2] - v0.v[2]) * amount));
            }
            static catmullrom(v0, v1, v2, v3, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                return new Vector3(0.5 * ((((2.0 * v1.v[0])
                    + ((-v0.v[0] + v2.v[0]) * amount))
                    + (((((2.0 * v0.v[0]) - (5.0 * v1.v[0]))
                        + (4.0 * v2.v[0])) - v3.v[0]) * n0))
                    + ((((-v0.v[0] + (3.0 * v1.v[0]))
                        - (3.0 * v2.v[0])) + v3.v[0]) * n1)), 0.5 * ((((2.0 * v1.v[1])
                    + ((-v0.v[1] + v2.v[1]) * amount))
                    + (((((2.0 * v0.v[1]) - (5.0 * v1.v[1]))
                        + (4.0 * v2.v[1])) - v3.v[1]) * n0))
                    + ((((-v0.v[1] + (3.0 * v1.v[1]))
                        - (3.0 * v2.v[1])) + v3.v[1]) * n1)), 0.5 * ((((2.0 * v1.v[2])
                    + ((-v0.v[2] + v2.v[2]) * amount))
                    + (((((2.0 * v0.v[2]) - (5.0 * v1.v[2]))
                        + (4.0 * v2.v[2])) - v3.v[2]) * n0))
                    + ((((-v0.v[2] + (3.0 * v1.v[2]))
                        - (3.0 * v2.v[2])) + v3.v[2]) * n1)));
            }
            static hermite(v0, t0, v1, t1, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                const n2 = ((2.0 * n1) - (3.0 * n0)) + 1.0;
                const n3 = (-2.0 * n1) + (3.0 * n0);
                const n4 = (n1 - (2.0 * n0)) + amount;
                const n5 = n1 - n0;
                return new Vector3((((v0.v[0] * n2) + (v1.v[0] * n3)) + (t0.v[0] * n4)) + (t1.v[0] * n5), (((v0.v[1] * n2) + (v1.v[1] * n3)) + (t0.v[1] * n4)) + (t1.v[1] * n5), (((v0.v[2] * n2) + (v1.v[2] * n3)) + (t0.v[2] * n4)) + (t1.v[2] * n5));
            }
            static transform(v0, m0) {
                return new Vector3((((v0.v[0] * m0.v[0]) + (v0.v[1] * m0.v[4])) + (v0.v[2] * m0.v[8])) + m0.v[12], (((v0.v[0] * m0.v[1]) + (v0.v[1] * m0.v[5])) + (v0.v[2] * m0.v[9])) + m0.v[13], (((v0.v[0] * m0.v[2]) + (v0.v[1] * m0.v[6])) + (v0.v[2] * m0.v[10])) + m0.v[14]);
            }
            static transformNormal(n0, m0) {
                return new Vector3(((n0.v[0] * m0.v[0]) + (n0.v[1] * m0.v[4])) + (n0.v[2] * m0.v[8]), ((n0.v[0] * m0.v[1]) + (n0.v[1] * m0.v[5])) + (n0.v[2] * m0.v[9]), ((n0.v[0] * m0.v[2]) + (n0.v[1] * m0.v[6])) + (n0.v[2] * m0.v[10]));
            }
            static transformQuaternion(v0, q0) {
                const n0 = q0.v[0] + q0.v[0];
                const n1 = q0.v[1] + q0.v[1];
                const n2 = q0.v[2] + q0.v[2];
                const n3 = q0.v[3] * n0;
                const n4 = q0.v[3] * n1;
                const n5 = q0.v[3] * n2;
                const n6 = q0.v[0] * n0;
                const n7 = q0.v[0] * n1;
                const n8 = q0.v[0] * n2;
                const n9 = q0.v[1] * n1;
                const n10 = q0.v[1] * n2;
                const n11 = q0.v[2] * n2;
                return new Vector3(((v0.v[0] * ((1.0 - n9) - n11)) + (v0.v[1] * (n7 - n5))) + (v0.v[2] * (n8 + n4)), ((v0.v[0] * (n7 + n5)) + (v0.v[1] * ((1.0 - n6) - n11))) + (v0.v[2] * (n10 - n3)), ((v0.v[0] * (n8 - n4)) + (v0.v[1] * (n10 + n3))) + (v0.v[2] * ((1.0 - n6) - n9)));
            }
            static add(v0, v1) {
                return new Vector3(v0.v[0] + v1.v[0], v0.v[1] + v1.v[1], v0.v[2] + v1.v[2]);
            }
            static sub(v0, v1) {
                return new Vector3(v0.v[0] - v1.v[0], v0.v[1] - v1.v[1], v0.v[2] - v1.v[2]);
            }
            static mul(v0, v1) {
                return new Vector3(v0.v[0] * v1.v[0], v0.v[1] * v1.v[1], v0.v[2] * v1.v[2]);
            }
            static div(v0, v1) {
                return new Vector3(v0.v[0] / v1.v[0], v0.v[1] / v1.v[1], v0.v[2] / v1.v[2]);
            }
            static scale(v0, scalar) {
                return new Vector3(v0.v[0] * scalar, v0.v[1] * scalar, v0.v[2] * scalar);
            }
            static negate(v0) {
                return new Vector3(-v0.v[0], -v0.v[1], -v0.v[2]);
            }
            static clone(v0) {
                return new Vector3(v0.v[0], v0.v[1], v0.v[2]);
            }
            static create(x, y, z) {
                return new Vector3(x, y, z);
            }
        }
        exports.Vector3 = Vector3;
        Vector3.MAX_VALUE = new Vector3(f32.max, f32.max, f32.max);
        Vector3.MIN_VALUE = new Vector3(f32.min, f32.min, f32.min);
    });
    define("engine/math/quaternion", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Quaternion {
            constructor(x, y, z, w) {
                this.v = new Float32Array(4);
                this.v[0] = x;
                this.v[1] = y;
                this.v[2] = z;
                this.v[3] = w;
            }
            toString() {
                return `[${this.v[0]}, ${this.v[1]}, ${this.v[2]}, ${this.v[3]}]`;
            }
            kind() {
                return 'Quaternion';
            }
            clone() {
                return Quaternion.clone(this);
            }
            get x() {
                return this.v[0];
            }
            get y() {
                return this.v[1];
            }
            get z() {
                return this.v[2];
            }
            get w() {
                return this.v[3];
            }
            set x(value) {
                this.v[0] = value;
            }
            set y(value) {
                this.v[1] = value;
            }
            set z(value) {
                this.v[2] = value;
            }
            set w(value) {
                this.v[3] = value;
            }
            length() {
                return Quaternion.getLength(this);
            }
            lengthSq() {
                return Quaternion.getLengthSq(this);
            }
            normalize() {
                return Quaternion.normalize(this);
            }
            dot(v0) {
                return Quaternion.dot(this, v0);
            }
            concat(q0) {
                return Quaternion.concat(this, q0);
            }
            add(q0) {
                return Quaternion.add(this, q0);
            }
            sub(q0) {
                return Quaternion.sub(this, q0);
            }
            mul(q0) {
                return Quaternion.mul(this, q0);
            }
            div(q0) {
                return Quaternion.div(this, q0);
            }
            static equals(q0, q1) {
                return (q0.v[0] === q1.v[0] &&
                    q0.v[1] === q1.v[1] &&
                    q0.v[2] === q1.v[2] &&
                    q0.v[3] === q1.v[3]);
            }
            static getLength(q0) {
                return Math.sqrt((q0.v[0] * q0.v[0]) +
                    (q0.v[1] * q0.v[1]) +
                    (q0.v[2] * q0.v[2]) +
                    (q0.v[3] * q0.v[3]));
            }
            static getLengthSq(q0) {
                return ((q0.v[0] * q0.v[0]) +
                    (q0.v[1] * q0.v[1]) +
                    (q0.v[2] * q0.v[2]) +
                    (q0.v[3] * q0.v[3]));
            }
            static normalize(q0) {
                const len = 1.0 / Math.sqrt((q0.v[0] * q0.v[0]) +
                    (q0.v[1] * q0.v[1]) +
                    (q0.v[2] * q0.v[2]) +
                    (q0.v[3] * q0.v[3]));
                return new Quaternion(q0.v[0] * len, q0.v[1] * len, q0.v[2] * len, q0.v[3] * len);
            }
            static dot(q0, q1) {
                return ((q0.v[0] * q1.v[0]) +
                    (q0.v[1] * q1.v[1]) +
                    (q0.v[2] * q1.v[2]) +
                    (q0.v[3] * q1.v[3]));
            }
            static conjugate(q0) {
                return new Quaternion(-q0.v[0], -q0.v[1], -q0.v[2], q0.v[3]);
            }
            static inverse(q0) {
                const n0 = (((q0.v[0] * q0.v[0]) +
                    (q0.v[1] * q0.v[1])) +
                    (q0.v[2] * q0.v[2])) +
                    (q0.v[3] * q0.v[3]);
                const n1 = 1.0 / n0;
                return new Quaternion(-q0.v[0] * n1, -q0.v[1] * n1, -q0.v[2] * n1, -q0.v[3] * n1);
            }
            static slerp(q0, q1, amount) {
                let n0 = 0.0;
                let n1 = 0.0;
                const n2 = amount;
                let n3 = (((q0.v[0] * q1.v[0]) +
                    (q0.v[1] * q1.v[1])) +
                    (q0.v[2] * q1.v[2])) +
                    (q0.v[3] * q1.v[3]);
                let flag = false;
                if (n3 < 0.0) {
                    flag = true;
                    n3 = -n3;
                }
                if (n3 > 0.999999) {
                    n1 = 1.0 - n2;
                    n0 = flag ? -n2 : n2;
                }
                else {
                    const n4 = Math.acos(n3);
                    const n5 = 1.0 / Math.sin(n4);
                    n1 = Math.sin(((1.0 - n2) * n4)) * n5;
                    n0 = flag ? (-Math.sin(n2 * n4) * n5) : (Math.sin(n2 * n4) * n5);
                }
                return new Quaternion((n1 * q0.v[0]) + (n0 * q1.v[0]), (n1 * q0.v[1]) + (n0 * q1.v[1]), (n1 * q0.v[2]) + (n0 * q1.v[2]), (n1 * q0.v[3]) + (n0 * q1.v[3]));
            }
            static lerp(q0, q1, amount) {
                const q2 = Quaternion.zero();
                const n0 = amount;
                const n1 = 1.0 - n0;
                const n2 = (((q0.v[0] * q1.v[0]) +
                    (q0.v[1] * q1.v[1])) +
                    (q0.v[2] * q1.v[2])) +
                    (q0.v[3] * q1.v[3]);
                if (n2 >= 0.0) {
                    q2.v[0] = (n1 * q0.v[0]) + (n0 * q1.v[0]);
                    q2.v[1] = (n1 * q0.v[1]) + (n0 * q1.v[1]);
                    q2.v[2] = (n1 * q0.v[2]) + (n0 * q1.v[2]);
                    q2.v[3] = (n1 * q0.v[3]) + (n0 * q1.v[3]);
                }
                else {
                    q2.v[0] = (n1 * q0.v[0]) - (n0 * q1.v[0]);
                    q2.v[1] = (n1 * q0.v[1]) - (n0 * q1.v[1]);
                    q2.v[2] = (n1 * q0.v[2]) - (n0 * q1.v[2]);
                    q2.v[3] = (n1 * q0.v[3]) - (n0 * q1.v[3]);
                }
                const n3 = (((q2.v[0] * q2.v[0]) +
                    (q2.v[1] * q2.v[1])) +
                    (q2.v[2] * q2.v[2])) +
                    (q2.v[3] * q2.v[3]);
                const n4 = 1.0 / Math.sqrt(n3);
                q2.v[0] *= n4;
                q2.v[1] *= n4;
                q2.v[2] *= n4;
                q2.v[3] *= n4;
                return q2;
            }
            static fromAxisAngle(v0, angle) {
                const n0 = angle * 0.5;
                const n1 = Math.sin(n0);
                const n2 = Math.cos(n0);
                return new Quaternion(v0.v[0] * n1, v0.v[1] * n1, v0.v[2] * n1, n2);
            }
            static fromMatrix(m0) {
                const n0 = (m0.v[0] + m0.v[5]) + m0.v[10];
                if (n0 > 0.0) {
                    const n1 = Math.sqrt(n0 + 1.0);
                    const n2 = 0.5 / n1;
                    return new Quaternion((m0.v[6] - m0.v[9]) * n2, (m0.v[8] - m0.v[2]) * n2, (m0.v[1] - m0.v[4]) * n2, n1 * 0.5);
                }
                else if ((m0.v[0] >= m0.v[5]) && (m0.v[0] >= m0.v[10])) {
                    const n1 = Math.sqrt(((1.0 + m0.v[0]) - m0.v[5]) - m0.v[10]);
                    const n2 = 0.5 / n1;
                    return new Quaternion(0.5 * n1, (m0.v[1] + m0.v[4]) * n2, (m0.v[2] + m0.v[8]) * n2, (m0.v[6] - m0.v[9]) * n2);
                }
                else if (m0.v[5] > m0.v[10]) {
                    const n1 = Math.sqrt(((1.0 + m0.v[5]) - m0.v[0]) - m0.v[10]);
                    const n2 = 0.5 / n1;
                    return new Quaternion((m0.v[4] + m0.v[1]) * n2, 0.5 * n1, (m0.v[9] + m0.v[6]) * n2, (m0.v[8] - m0.v[2]) * n2);
                }
                else {
                    const n1 = Math.sqrt(((1.0 + m0.v[10]) - m0.v[0]) - m0.v[5]);
                    const n2 = 0.5 / n1;
                    return new Quaternion((m0.v[8] + m0.v[2]) * n2, (m0.v[9] + m0.v[6]) * n2, 0.5 * n1, (m0.v[1] - m0.v[4]) * n2);
                }
            }
            static concat(q0, q1) {
                const n0 = q1.v[0];
                const n1 = q1.v[1];
                const n2 = q1.v[2];
                const n3 = q1.v[3];
                const n4 = q0.v[0];
                const n5 = q0.v[1];
                const n6 = q0.v[2];
                const n7 = q0.v[3];
                const n8 = (n1 * n6) - (n2 * n5);
                const n9 = (n2 * n4) - (n0 * n6);
                const n10 = (n0 * n5) - (n1 * n4);
                const n11 = ((n0 * n4) + (n1 * n5)) + (n2 * n6);
                return new Quaternion(((n0 * n7) + (n4 * n3)) + n8, ((n1 * n7) + (n5 * n3)) + n9, ((n2 * n7) + (n6 * n3)) + n10, (n3 * n7) - n11);
            }
            static add(q0, q1) {
                return new Quaternion(q0.v[0] + q1.v[0], q0.v[1] + q1.v[1], q0.v[2] + q1.v[2], q0.v[3] + q1.v[3]);
            }
            static sub(q0, q1) {
                return new Quaternion(q0.v[0] - q1.v[0], q0.v[1] - q1.v[1], q0.v[2] - q1.v[2], q0.v[3] - q1.v[3]);
            }
            static mul(q0, q1) {
                const n0 = q0.v[0];
                const n1 = q0.v[1];
                const n2 = q0.v[2];
                const n3 = q0.v[3];
                const n4 = q1.v[0];
                const n5 = q1.v[1];
                const n6 = q1.v[2];
                const n7 = q1.v[3];
                const n8 = (n1 * n6) - (n2 * n5);
                const n9 = (n2 * n4) - (n0 * n6);
                const n10 = (n0 * n5) - (n1 * n4);
                const n11 = ((n0 * n4) + (n1 * n5)) + (n2 * n6);
                return new Quaternion(((n0 * n7) + (n4 * n3)) + n8, ((n1 * n7) + (n5 * n3)) + n9, ((n2 * n7) + (n6 * n3)) + n10, (n3 * n7) - n11);
            }
            static div(q0, q1) {
                const n0 = q0.v[0];
                const n1 = q0.v[1];
                const n2 = q0.v[2];
                const n3 = q0.v[3];
                const n4 = (((q1.v[0] * q1.v[0]) +
                    (q1.v[1] * q1.v[1])) +
                    (q1.v[2] * q1.v[2])) +
                    (q1.v[3] * q1.v[3]);
                const n5 = 1.0 / n4;
                const n6 = -q1.v[0] * n5;
                const n7 = -q1.v[1] * n5;
                const n8 = -q1.v[2] * n5;
                const n9 = q1.v[3] * n5;
                const n10 = (n1 * n8) - (n2 * n7);
                const n11 = (n2 * n6) - (n0 * n8);
                const n12 = (n0 * n7) - (n1 * n6);
                const n13 = ((n0 * n6) + (n1 * n7)) + (n2 * n8);
                return new Quaternion(((n0 * n9) + (n6 * n3)) + n10, ((n1 * n9) + (n7 * n3)) + n11, ((n2 * n9) + (n8 * n3)) + n12, (n3 * n9) - n13);
            }
            static negate(q0) {
                return new Quaternion(-q0.v[0], -q0.v[1], -q0.v[2], -q0.v[3]);
            }
            static clone(q0) {
                return new Quaternion(q0.v[0], q0.v[1], q0.v[2], q0.v[3]);
            }
            static zero() {
                return new Quaternion(0, 0, 0, 0);
            }
            static create(x, y, z, w) {
                return new Quaternion(x, y, z, w);
            }
        }
        exports.Quaternion = Quaternion;
    });
    define("engine/math/vector4", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const f32 = { min: Number.MIN_VALUE, max: Number.MAX_VALUE };
        class Vector4 {
            constructor(x, y, z, w) {
                this.v = new Float32Array(4);
                this.v[0] = x;
                this.v[1] = y;
                this.v[2] = z;
                this.v[3] = w;
            }
            toString() {
                return `[${this.v[0]}, ${this.v[1]}, ${this.v[2]}, ${this.v[3]}]`;
            }
            kind() {
                return 'Vector4';
            }
            clone() {
                return Vector4.clone(this);
            }
            get x() {
                return this.v[0];
            }
            get y() {
                return this.v[1];
            }
            get z() {
                return this.v[2];
            }
            get w() {
                return this.v[3];
            }
            set x(value) {
                this.v[0] = value;
            }
            set y(value) {
                this.v[1] = value;
            }
            set z(value) {
                this.v[2] = value;
            }
            set w(value) {
                this.v[3] = value;
            }
            length() {
                return Vector4.getLength(this);
            }
            lengthSq() {
                return Vector4.getLengthSq(this);
            }
            normalize() {
                return Vector4.normalize(this);
            }
            dot(v0) {
                return Vector4.dot(this, v0);
            }
            add(v0) {
                return Vector4.add(this, v0);
            }
            sub(v0) {
                return Vector4.sub(this, v0);
            }
            mul(v0) {
                return Vector4.mul(this, v0);
            }
            div(v0) {
                return Vector4.div(this, v0);
            }
            scale(s0) {
                return Vector4.scale(this, s0);
            }
            negate() {
                return Vector4.negate(this);
            }
            static zero() {
                return new Vector4(0.0, 0.0, 0.0, 0.0);
            }
            static one() {
                return new Vector4(1.0, 1.0, 1.0, 1.0);
            }
            static left() {
                return new Vector4(-1.0, 0.0, 0.0, 0.0);
            }
            static unitX() {
                return new Vector4(1.0, 0.0, 0.0, 0.0);
            }
            static unitY() {
                return new Vector4(0.0, 1.0, 0.0, 0.0);
            }
            static unitZ() {
                return new Vector4(0.0, 0.0, 1.0, 0.0);
            }
            static unitW() {
                return new Vector4(0.0, 0.0, 0.0, 1.0);
            }
            static equals(v0, v1) {
                return (v0.v[0] === v1.v[0] &&
                    v0.v[1] === v1.v[1] &&
                    v0.v[2] === v1.v[2] &&
                    v0.v[3] === v1.v[3]);
            }
            static getLength(v0) {
                return Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]) +
                    (v0.v[3] * v0.v[3]));
            }
            static getLengthSq(v0) {
                return ((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]) +
                    (v0.v[3] * v0.v[3]));
            }
            static distance(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                const z = v0.v[2] - v1.v[2];
                const w = v0.v[3] - v1.v[3];
                return Math.sqrt((x * x) + (y * y) + (z * z) + (w * w));
            }
            static distanceSq(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                const z = v0.v[2] - v1.v[2];
                const w = v0.v[3] - v1.v[3];
                return ((x * x) + (y * y) + (z * z) + (w * w));
            }
            static dot(v0, v1) {
                return ((v0.v[0] * v1.v[0]) +
                    (v0.v[1] * v1.v[1]) +
                    (v0.v[2] * v1.v[2]) +
                    (v0.v[3] * v1.v[3]));
            }
            static normalize(v0) {
                const len = 1.0 / Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]) +
                    (v0.v[2] * v0.v[2]) +
                    (v0.v[3] * v0.v[3]));
                return new Vector4(v0.v[0] * len, v0.v[1] * len, v0.v[2] * len, v0.v[3] * len);
            }
            static abs(v0) {
                return new Vector4(Math.abs(v0.v[0]), Math.abs(v0.v[1]), Math.abs(v0.v[2]), Math.abs(v0.v[3]));
            }
            static min(v0, v1) {
                return new Vector4((v0.v[0] < v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] < v1.v[1]) ? v0.v[1] : v1.v[1], (v0.v[2] < v1.v[2]) ? v0.v[2] : v1.v[2], (v0.v[3] < v1.v[3]) ? v0.v[3] : v1.v[3]);
            }
            static max(v0, v1) {
                return new Vector4((v0.v[0] > v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] > v1.v[1]) ? v0.v[1] : v1.v[1], (v0.v[2] > v1.v[2]) ? v0.v[2] : v1.v[2], (v0.v[3] > v1.v[3]) ? v0.v[3] : v1.v[3]);
            }
            static clamp(v0, min, max) {
                let x = v0.v[0];
                let y = v0.v[1];
                let z = v0.v[2];
                let w = v0.v[3];
                x = (x > max.v[0]) ? max.v[0] : x;
                x = (x < min.v[0]) ? min.v[0] : x;
                y = (y > max.v[1]) ? max.v[1] : y;
                y = (y < min.v[1]) ? min.v[1] : y;
                z = (z > max.v[2]) ? max.v[2] : z;
                z = (z < min.v[2]) ? min.v[2] : z;
                w = (w > max.v[3]) ? max.v[3] : w;
                w = (w < min.v[3]) ? min.v[3] : w;
                return new Vector4(x, y, z, w);
            }
            static lerp(v0, v1, amount) {
                return new Vector4(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount), v0.v[2] + ((v1.v[2] - v0.v[2]) * amount), v0.v[3] + ((v1.v[3] - v0.v[3]) * amount));
            }
            static barycentric(v0, v1, v2, amount0, amount1) {
                return new Vector4((v0.v[0] + (amount0 * (v1.v[0] - v0.v[0]))) + (amount1 * (v2.v[0] - v0.v[0])), (v0.v[1] + (amount0 * (v1.v[1] - v0.v[1]))) + (amount1 * (v2.v[1] - v0.v[1])), (v0.v[2] + (amount0 * (v1.v[2] - v0.v[2]))) + (amount1 * (v2.v[2] - v0.v[2])), (v0.v[3] + (amount0 * (v1.v[3] - v0.v[3]))) + (amount1 * (v2.v[3] - v0.v[3])));
            }
            static smoothstep(v0, v1, amount) {
                amount = (amount > 1.0) ? 1.0 : ((amount < 0.0) ? 0.0 : amount);
                amount = (amount * amount) * (3.0 - (2.0 * amount));
                return new Vector4(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount), v0.v[2] + ((v1.v[2] - v0.v[2]) * amount), v0.v[3] + ((v1.v[3] - v0.v[3]) * amount));
            }
            static catmullrom(v0, v1, v2, v3, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                return new Vector4(0.5 * ((((2.0 * v1.v[0])
                    + ((-v0.v[0] + v2.v[0]) * amount))
                    + (((((2.0 * v0.v[0]) - (5.0 * v1.v[0]))
                        + (4.0 * v2.v[0])) - v3.v[0]) * n0))
                    + ((((-v0.v[0] + (3.0 * v1.v[0]))
                        - (3.0 * v2.v[0])) + v3.v[0]) * n1)), 0.5 * ((((2.0 * v1.v[1])
                    + ((-v0.v[1] + v2.v[1]) * amount))
                    + (((((2.0 * v0.v[1]) - (5.0 * v1.v[1]))
                        + (4.0 * v2.v[1])) - v3.v[1]) * n0))
                    + ((((-v0.v[1] + (3.0 * v1.v[1]))
                        - (3.0 * v2.v[1])) + v3.v[1]) * n1)), 0.5 * ((((2.0 * v1.v[2])
                    + ((-v0.v[2] + v2.v[2]) * amount))
                    + (((((2.0 * v0.v[2]) - (5.0 * v1.v[2]))
                        + (4.0 * v2.v[2])) - v3.v[2]) * n0))
                    + ((((-v0.v[2] + (3.0 * v1.v[2]))
                        - (3.0 * v2.v[2])) + v3.v[2]) * n1)), 0.5 * ((((2.0 * v1.v[3])
                    + ((-v0.v[3] + v2.v[3]) * amount))
                    + (((((2.0 * v0.v[3]) - (5.0 * v1.v[3]))
                        + (4.0 * v2.v[3])) - v3.v[3]) * n0))
                    + ((((-v0.v[3] + (3.0 * v1.v[3]))
                        - (3.0 * v2.v[3])) + v3.v[3]) * n1)));
            }
            static hermite(v0, t0, v1, t1, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                const n2 = ((2.0 * n1) - (3.0 * n0)) + 1.0;
                const n3 = (-2.0 * n1) + (3.0 * n0);
                const n4 = (n1 - (2.0 * n0)) + amount;
                const n5 = n1 - n0;
                return new Vector4((((v0.v[0] * n2) + (v1.v[0] * n3)) + (t0.v[0] * n4)) + (t1.v[0] * n5), (((v0.v[1] * n2) + (v1.v[1] * n3)) + (t0.v[1] * n4)) + (t1.v[1] * n5), (((v0.v[2] * n2) + (v1.v[2] * n3)) + (t0.v[2] * n4)) + (t1.v[2] * n5), (((v0.v[3] * n2) + (v1.v[3] * n3)) + (t0.v[3] * n4)) + (t1.v[3] * n5));
            }
            static fast_transform(v0, m0, out) {
                out.v[0] = (((v0.v[0] * m0.v[0]) + (v0.v[1] * m0.v[4])) + (v0.v[2] * m0.v[8])) + (v0.v[3] * m0.v[12]),
                    out.v[1] = (((v0.v[0] * m0.v[1]) + (v0.v[1] * m0.v[5])) + (v0.v[2] * m0.v[9])) + (v0.v[3] * m0.v[13]),
                    out.v[2] = (((v0.v[0] * m0.v[2]) + (v0.v[1] * m0.v[6])) + (v0.v[2] * m0.v[10])) + (v0.v[3] * m0.v[14]),
                    out.v[3] = (((v0.v[0] * m0.v[3]) + (v0.v[1] * m0.v[7])) + (v0.v[2] * m0.v[11])) + (v0.v[3] * m0.v[15]);
            }
            static transform(v0, m0) {
                const out = Vector4.zero();
                Vector4.fast_transform(v0, m0, out);
                return out;
            }
            static transformQuaternion(v0, q0) {
                const n0 = q0.v[0] + q0.v[0];
                const n1 = q0.v[1] + q0.v[1];
                const n2 = q0.v[2] + q0.v[2];
                const n3 = q0.v[3] * n0;
                const n4 = q0.v[3] * n1;
                const n5 = q0.v[3] * n2;
                const n6 = q0.v[0] * n0;
                const n7 = q0.v[0] * n1;
                const n8 = q0.v[0] * n2;
                const n9 = q0.v[1] * n1;
                const n10 = q0.v[1] * n2;
                const n11 = q0.v[2] * n2;
                return new Vector4((v0.v[0] * ((1.0 - n9) - n11)) + (v0.v[1] * (n7 - n5)), (v0.v[0] * (n7 + n5)) + (v0.v[1] * ((1.0 - n6) - n11)), (v0.v[0] * (n8 - n4)) + (v0.v[1] * (n10 + n3)), 0.0);
            }
            static add(v0, v1) {
                return new Vector4(v0.v[0] + v1.v[0], v0.v[1] + v1.v[1], v0.v[2] + v1.v[2], v0.v[3] + v1.v[3]);
            }
            static sub(v0, v1) {
                return new Vector4(v0.v[0] - v1.v[0], v0.v[1] - v1.v[1], v0.v[2] - v1.v[2], v0.v[3] - v1.v[3]);
            }
            static mul(v0, v1) {
                return new Vector4(v0.v[0] * v1.v[0], v0.v[1] * v1.v[1], v0.v[2] * v1.v[2], v0.v[3] * v1.v[3]);
            }
            static div(v0, v1) {
                return new Vector4(v0.v[0] / v1.v[0], v0.v[1] / v1.v[1], v0.v[2] / v1.v[2], v0.v[3] / v1.v[3]);
            }
            static scale(v0, scalar) {
                return new Vector4(v0.v[0] * scalar, v0.v[1] * scalar, v0.v[2] * scalar, v0.v[3] * scalar);
            }
            static negate(v0) {
                return new Vector4(-v0.v[0], -v0.v[1], -v0.v[2], -v0.v[3]);
            }
            static clone(v0) {
                return new Vector4(v0.v[0], v0.v[1], v0.v[2], v0.v[3]);
            }
            static create(x, y, z, w) {
                return new Vector4(x, y, z, w);
            }
        }
        exports.Vector4 = Vector4;
        Vector4.MAX_VALUE = new Vector4(f32.max, f32.max, f32.max, f32.max);
        Vector4.MIN_VALUE = new Vector4(f32.min, f32.min, f32.min, f32.min);
    });
    define("engine/math/single", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Single {
            constructor(x) {
                this.v = new Float32Array(1);
                this.v[0] = x;
            }
            toString() {
                return `${this.v[0]}`;
            }
            kind() {
                return "Single";
            }
            clone() {
                return Single.clone(this);
            }
            get x() {
                return this.v[0];
            }
            set x(value) {
                this.v[0] = value;
            }
            negate() {
                return Single.negate(this);
            }
            add(s0) {
                return Single.add(this, s0);
            }
            sub(s0) {
                return Single.sub(this, s0);
            }
            mul(s0) {
                return Single.mul(this, s0);
            }
            div(s0) {
                return Single.div(this, s0);
            }
            static distance(s0, s1) {
                return Math.abs(s0.v[0] - s1.v[0]);
            }
            static barycentric(s0, s1, s2, amount0, amount1) {
                return new Single((s0.v[0] + (amount0 * (s1.v[0] - s0.v[0]))) + (amount0 * (s2.v[0] - s0.v[0])));
            }
            static catmullrom(s0, s1, s2, s3, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                return new Single((0.5 * ((((2.0 * s1.v[0]) + ((-s0.v[0] + s2.v[0]) * amount)) +
                    (((((2.0 * s0.v[0]) - (5.0 * s1.v[0])) +
                        (4.0 * s2.v[0])) - s3.v[0]) * n0)) +
                    ((((-s0.v[0] + (3.0 * s1.v[0])) -
                        (3.0 * s2.v[0])) + s3.v[0]) * n1))));
            }
            static hermite(s0, t0, s1, t1, amount) {
                const n0 = amount;
                const n1 = n0 * n0;
                const n2 = n0 * n1;
                const n3 = ((2.0 * n2) - (3.0 * n1)) + 1.0;
                const n4 = (-2.0 * n2) + (3.0 * n1);
                const n5 = (n2 - (2.0 * n1)) + n0;
                const n6 = n2 - n1;
                return new Single((((s0.v[0] * n3) + (s1.v[0] * n4)) + (t0.v[0] * n5)) + (t1.v[0] * n6));
            }
            static lerp(s0, s1, amount) {
                return new Single(s0.v[0] + ((s1.v[0] - s0.v[0]) * amount));
            }
            static smoothstep(value1, value2, amount) {
                const num = Single.clamp(new Single(amount), 0.0, 1.0).x;
                return Single.lerp(value1, value2, (num * num) * (3.0 - (2.0 * num)));
            }
            static clamp(s0, min, max) {
                const n0 = (s0.v[0] > max) ? max : s0.v[0];
                const n1 = (n0 < max) ? min : n0;
                return new Single(n1);
            }
            static abs(s0) {
                return new Single(Math.abs(s0.v[0]));
            }
            static min(s0, s1) {
                return (s0.v[0] < s1.v[0])
                    ? new Single(s0.v[0])
                    : new Single(s1.v[0]);
            }
            static max(s0, s1) {
                return (s0.v[0] > s1.v[0])
                    ? new Single(s0.v[0])
                    : new Single(s1.v[0]);
            }
            static negate(s0) {
                return new Single(-s0.v[0]);
            }
            static add(s0, s1) {
                return new Single(s0.v[0] + s1.v[0]);
            }
            static sub(s0, s1) {
                return new Single(s0.v[0] - s1.v[0]);
            }
            static mul(s0, s1) {
                return new Single(s0.v[0] * s1.v[0]);
            }
            static div(s0, s1) {
                return new Single(s0.v[0] / s1.v[0]);
            }
            static clone(s0) {
                return new Single(s0.v[0]);
            }
            static create(x) {
                return new Single(x);
            }
        }
        exports.Single = Single;
    });
    define("engine/math/sphere", ["require", "exports", "engine/math/single", "engine/math/vector3"], function (require, exports, single_1, vector3_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Sphere {
            constructor(position, radius) {
                this.position = position;
                this.radius = radius;
            }
            toString() {
                return `{ position: ${this.position.toString()}, radius: ${this.radius}}`;
            }
            kind() {
                return 'Sphere';
            }
            clone() {
                return Sphere.clone(this);
            }
            static equals(s0, s1) {
                return (vector3_1.Vector3.equals(s0.position, s1.position) &&
                    s0.radius === s1.radius);
            }
            static merge(s0, s1) {
                let n0 = vector3_1.Vector3.sub(s1.position, s0.position);
                let n1 = n0.length();
                let n2 = s0.radius;
                let n3 = s1.radius;
                if ((n2 + n3) >= n1) {
                    if ((n2 - n3) >= n1) {
                        return s0.clone();
                    }
                    if ((n3 - n2) >= n1) {
                        return s1.clone();
                    }
                }
                let n4 = n0.scale(1.0 / n1);
                let n5 = single_1.Single.min(new single_1.Single(-n2), new single_1.Single(n1 - n3));
                let n6 = single_1.Single.sub(single_1.Single.max(new single_1.Single(n2), new single_1.Single(n1 + n3)), n5)
                    .mul(new single_1.Single(0.5));
                return new Sphere(s0.position.add(n4.scale(n6.add(n5).x)), n6.x);
            }
            static fromBox(b0) {
                let center = vector3_1.Vector3.lerp(b0.min, b0.max, 0.5);
                let distance = vector3_1.Vector3.distance(b0.min, b0.max);
                return new Sphere(center, distance * 0.5);
            }
            static fromPoints(points) {
                let radius = 0.0;
                let center = vector3_1.Vector3.zero();
                let n0 = vector3_1.Vector3.zero();
                let n1 = vector3_1.Vector3.zero();
                let n2 = vector3_1.Vector3.zero();
                let n3 = vector3_1.Vector3.zero();
                let n4 = vector3_1.Vector3.zero();
                let n5 = vector3_1.Vector3.zero();
                for (let i = 0; i < points.length; i++) {
                    let p = points[i];
                    if (p.v[0] < n0.v[0]) {
                        n0 = p.clone();
                    }
                    if (p.v[0] > n1.v[0]) {
                        n1 = p.clone();
                    }
                    if (p.v[1] < n2.v[1]) {
                        n2 = p.clone();
                    }
                    if (p.v[1] > n3.v[1]) {
                        n3 = p.clone();
                    }
                    if (p.v[2] < n4.v[2]) {
                        n4 = p.clone();
                    }
                    if (p.v[2] > n5.v[2]) {
                        n5 = p.clone();
                    }
                }
                let n6 = vector3_1.Vector3.distance(n1, n0);
                let n7 = vector3_1.Vector3.distance(n3, n2);
                let n8 = vector3_1.Vector3.distance(n5, n4);
                if (n6 > n7) {
                    if (n6 > n8) {
                        center = vector3_1.Vector3.lerp(n1, n0, 0.5);
                        radius = n6 * 0.5;
                    }
                    else {
                        center = vector3_1.Vector3.lerp(n5, n4, 0.5);
                        radius = n8 * 0.5;
                    }
                }
                else if (n7 > n8) {
                    center = vector3_1.Vector3.lerp(n3, n2, 0.5);
                    radius = n7 * 0.5;
                }
                else {
                    center = vector3_1.Vector3.lerp(n5, n4, 0.5);
                    radius = n8 * 0.5;
                }
                for (let i = 0; i < points.length; i++) {
                    let v0 = points[i];
                    let v1 = new vector3_1.Vector3(v0.v[0] - center.v[0], v0.v[1] - center.v[1], v0.v[2] - center.v[2]);
                    let num3 = v1.length();
                    if (num3 > radius) {
                        radius = (radius + num3) * 0.5;
                        center = center.add(vector3_1.Vector3.scale(v1, 1.0 - (radius / num3)));
                    }
                }
                return new Sphere(center, radius);
            }
            static clone(s0) {
                return new Sphere(s0.position.clone(), s0.radius);
            }
            static create(center, radius) {
                return new Sphere(center, radius);
            }
        }
        exports.Sphere = Sphere;
    });
    define("engine/math/frustum", ["require", "exports", "engine/math/vector3", "engine/math/plane", "engine/math/ray"], function (require, exports, vector3_2, plane_1, ray_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const computeIntersectionRay = (p0, p1) => {
            const v0 = vector3_2.Vector3.cross(p0.normal(), p1.normal());
            const n0 = v0.lengthSq();
            const v1 = vector3_2.Vector3.scale(p1.normal(), -p0.d);
            const v2 = vector3_2.Vector3.scale(p0.normal(), p1.d);
            const v3 = vector3_2.Vector3.add(v1, v2);
            const v4 = vector3_2.Vector3.cross(v3, v0);
            const v5 = new vector3_2.Vector3(v4.v[0] / n0, v4.v[1] / n0, v4.v[2] / n0);
            return new ray_1.Ray(v5, v0);
        };
        const computeIntersectionVector = (plane, ray) => {
            const num = (-plane.d -
                vector3_2.Vector3.dot(plane.normal(), ray.position)) /
                vector3_2.Vector3.dot(plane.normal(), ray.direction);
            return vector3_2.Vector3.add(ray.position, vector3_2.Vector3.scale(ray.direction, num));
        };
        class Frustum {
            constructor(Matrix) {
                this.Matrix = Matrix;
                this.planes = Array.from({ length: 6 });
                this.planes[0] = new plane_1.Plane(-this.Matrix.v[2], -this.Matrix.v[6], -this.Matrix.v[10], -this.Matrix.v[14]);
                this.planes[1] = new plane_1.Plane(-this.Matrix.v[3] + this.Matrix.v[2], -this.Matrix.v[7] + this.Matrix.v[6], -this.Matrix.v[11] + this.Matrix.v[10], -this.Matrix.v[15] + this.Matrix.v[14]);
                this.planes[2] = new plane_1.Plane(-this.Matrix.v[3] - this.Matrix.v[0], -this.Matrix.v[7] - this.Matrix.v[4], -this.Matrix.v[11] - this.Matrix.v[8], -this.Matrix.v[15] - this.Matrix.v[12]);
                this.planes[3] = new plane_1.Plane(-this.Matrix.v[3] + this.Matrix.v[0], -this.Matrix.v[7] + this.Matrix.v[4], -this.Matrix.v[11] + this.Matrix.v[8], -this.Matrix.v[15] + this.Matrix.v[12]);
                this.planes[4] = new plane_1.Plane(-this.Matrix.v[3] + this.Matrix.v[1], -this.Matrix.v[7] + this.Matrix.v[5], -this.Matrix.v[11] + this.Matrix.v[9], -this.Matrix.v[15] + this.Matrix.v[13]);
                this.planes[5] = new plane_1.Plane(-this.Matrix.v[3] - this.Matrix.v[1], -this.Matrix.v[7] - this.Matrix.v[5], -this.Matrix.v[11] - this.Matrix.v[9], -this.Matrix.v[15] - this.Matrix.v[13]);
                for (let i = 0; i < this.planes.length; i++) {
                    const len = this.planes[i].normal().length();
                    this.planes[i].v[0] = this.planes[i].v[0] / len;
                    this.planes[i].v[1] = this.planes[i].v[1] / len;
                    this.planes[i].v[2] = this.planes[i].v[2] / len;
                    this.planes[i].v[3] = this.planes[i].v[3] / len;
                }
                this.corners = Array.from({ length: 8 });
                let ray = computeIntersectionRay(this.planes[0], this.planes[2]);
                this.corners[0] = computeIntersectionVector(this.planes[4], ray);
                this.corners[3] = computeIntersectionVector(this.planes[5], ray);
                ray = computeIntersectionRay(this.planes[3], this.planes[0]);
                this.corners[1] = computeIntersectionVector(this.planes[4], ray);
                this.corners[2] = computeIntersectionVector(this.planes[5], ray);
                ray = computeIntersectionRay(this.planes[2], this.planes[1]);
                this.corners[4] = computeIntersectionVector(this.planes[4], ray);
                this.corners[7] = computeIntersectionVector(this.planes[5], ray);
                ray = computeIntersectionRay(this.planes[1], this.planes[3]);
                this.corners[5] = computeIntersectionVector(this.planes[4], ray);
                this.corners[6] = computeIntersectionVector(this.planes[5], ray);
            }
            toString() {
                const buffer = [];
                buffer.push('{');
                buffer.push(`  planes: {`);
                buffer.push(`    near  : ${this.planes[0].toString()},`);
                buffer.push(`    far   : ${this.planes[1].toString()},`);
                buffer.push(`    left  : ${this.planes[2].toString()},`);
                buffer.push(`    right : ${this.planes[3].toString()},`);
                buffer.push(`    top   : ${this.planes[4].toString()},`);
                buffer.push(`    bottom: ${this.planes[5].toString()}`);
                buffer.push(`  },`);
                buffer.push(`  corners: [`);
                buffer.push(`    ${this.corners[0].toString()},`);
                buffer.push(`    ${this.corners[1].toString()},`);
                buffer.push(`    ${this.corners[2].toString()},`);
                buffer.push(`    ${this.corners[3].toString()},`);
                buffer.push(`    ${this.corners[4].toString()},`);
                buffer.push(`    ${this.corners[5].toString()},`);
                buffer.push(`    ${this.corners[6].toString()},`);
                buffer.push(`    ${this.corners[7].toString()}`);
                buffer.push(`  ]`);
                buffer.push('}');
                return buffer.join('\n');
            }
            kind() {
                return 'Frustum';
            }
            clone() {
                return Frustum.clone(this);
            }
            static equals(f0, f1) {
                return (f0.Matrix === f1.Matrix);
            }
            get near() {
                return this.planes[0];
            }
            get far() {
                return this.planes[1];
            }
            get left() {
                return this.planes[2];
            }
            get right() {
                return this.planes[3];
            }
            get top() {
                return this.planes[4];
            }
            get bottom() {
                return this.planes[5];
            }
            static clone(f0) {
                return new Frustum(f0.Matrix.clone());
            }
            static create(matrix) {
                return new Frustum(matrix);
            }
        }
        exports.Frustum = Frustum;
    });
    define("engine/math/triangle", ["require", "exports", "engine/math/vector3", "engine/math/plane"], function (require, exports, vector3_3, plane_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Triangle {
            constructor(v0, v1, v2) {
                this.v0 = v0;
                this.v1 = v1;
                this.v2 = v2;
            }
            toString() {
                return `{v0: ${this.v0.toString()}, v1: ${this.v0.toString()}, v2: ${this.v0.toString()}}`;
            }
            kind() {
                return 'Triangle';
            }
            clone() {
                return Triangle.clone(this);
            }
            plane() {
                return plane_2.Plane.fromPoints(this.v0, this.v1, this.v2);
            }
            static equals(t0, t1) {
                return (vector3_3.Vector3.equals(t0.v0, t1.v0) &&
                    vector3_3.Vector3.equals(t0.v1, t1.v1) &&
                    vector3_3.Vector3.equals(t0.v2, t1.v2));
            }
            static clone(t0) {
                return new Triangle(t0.v0, t0.v1, t0.v2);
            }
            static create(v0, v1, v2) {
                return new Triangle(v0, v1, v2);
            }
        }
        exports.Triangle = Triangle;
    });
    define("engine/math/ray", ["require", "exports", "engine/math/vector3"], function (require, exports, vector3_4) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const f32 = { max: Number.MAX_VALUE, min: Number.MIN_VALUE };
        const min = (a, b) => a < b ? a : b;
        const max = (a, b) => a > b ? a : b;
        class Ray {
            constructor(position, direction) {
                this.position = position;
                this.direction = direction;
            }
            toString() {
                return `{ position: ${this.position.toString()}, direction: ${this.direction.toString()} }`;
            }
            kind() {
                return 'Ray';
            }
            static equals(r0, r1) {
                return (vector3_4.Vector3.equals(r0.position, r1.position) &&
                    vector3_4.Vector3.equals(r0.direction, r1.direction));
            }
            static intersectPlane(ray, plane) {
                const n0 = ((plane.v[0] * ray.direction.v[0]) +
                    (plane.v[1] * ray.direction.v[1])) +
                    (plane.v[2] * ray.direction.v[2]);
                if (Math.abs(n0) < 1E-05) {
                    return undefined;
                }
                const n1 = ((plane.v[0] * ray.position.v[0]) +
                    (plane.v[1] * ray.position.v[1])) +
                    (plane.v[2] * ray.position.v[2]);
                let n2 = (-plane.v[3] - n1) / n0;
                if (n2 < 0.0) {
                    if (n2 < -1E-05) {
                        return undefined;
                    }
                    n2 = 0.0;
                }
                return n2;
            }
            static intersectTriangle(ray, triangle) {
                let result = undefined;
                const v0 = vector3_4.Vector3.sub(triangle.v1, triangle.v0);
                const v1 = vector3_4.Vector3.sub(triangle.v2, triangle.v0);
                const v2 = vector3_4.Vector3.cross(ray.direction, v1);
                const n0 = vector3_4.Vector3.dot(v0, v2);
                if (n0 > -0.00001) {
                    return undefined;
                }
                const n1 = 1.0 / n0;
                const v3 = vector3_4.Vector3.sub(ray.position, triangle.v0);
                const n2 = vector3_4.Vector3.dot(v3, v2) * n1;
                if (n2 < -0.001 || n2 > 1.001) {
                    return undefined;
                }
                const v4 = vector3_4.Vector3.cross(v3, v0);
                const n3 = vector3_4.Vector3.dot(ray.direction, v4) * n1;
                if (n3 < -0.001 || n2 + n3 > 1.001) {
                    return undefined;
                }
                result = vector3_4.Vector3.dot(v1, v4) * n1;
                if (result <= 0) {
                    return undefined;
                }
                return result;
            }
            static intersectBox(r0, b0) {
                let result = 0.0;
                let maxValue = f32.max;
                if (Math.abs(r0.direction.v[0]) < 1E-06) {
                    if ((r0.position.v[0] < b0.min.v[0]) ||
                        (r0.position.v[0] > b0.max.v[0])) {
                        return undefined;
                    }
                }
                else {
                    const n0 = 1.0 / r0.direction.v[0];
                    let n1 = (b0.min.v[0] - r0.position.v[0]) * n0;
                    let n2 = (b0.max.v[0] - r0.position.v[0]) * n0;
                    if (n1 > n2) {
                        const n3 = n1;
                        n1 = n2;
                        n2 = n3;
                    }
                    result = max(n1, result);
                    maxValue = min(n2, result);
                    if (result > maxValue) {
                        return undefined;
                    }
                }
                if (Math.abs(r0.direction.v[1]) < 1E-06) {
                    if ((r0.position.v[1] < b0.min.v[1]) ||
                        (r0.position.v[1] > b0.max.v[1])) {
                        return undefined;
                    }
                }
                else {
                    const n0 = 1.0 / r0.direction.v[1];
                    let n1 = (b0.min.v[1] - r0.position.v[1]) * n0;
                    let n2 = (b0.max.v[1] - r0.position.v[1]) * n0;
                    if (n1 > n2) {
                        const n3 = n1;
                        n1 = n2;
                        n2 = n3;
                    }
                    result = max(n1, result);
                    maxValue = min(n2, maxValue);
                    if (result > maxValue) {
                        return undefined;
                    }
                }
                if (Math.abs(r0.direction.v[2]) < 1E-06) {
                    if ((r0.position.v[2] < b0.min.v[2]) ||
                        (r0.position.v[2] > b0.max.v[2])) {
                        return undefined;
                    }
                }
                else {
                    const n0 = 1.0 / r0.direction.v[2];
                    let n1 = (b0.min.v[2] - r0.position.v[2]) * n0;
                    let n2 = (b0.max.v[2] - r0.position.v[2]) * n0;
                    if (n1 > n2) {
                        const n3 = n1;
                        n1 = n2;
                        n2 = n3;
                    }
                    result = max(n1, result);
                    maxValue = min(n2, maxValue);
                    if (result > maxValue) {
                        return undefined;
                    }
                }
                return result;
            }
            static intersectSphere(r0, s0) {
                const n0 = s0.position.v[0] - r0.position.v[0];
                const n1 = s0.position.v[1] - r0.position.v[1];
                const n2 = s0.position.v[2] - r0.position.v[2];
                const n3 = ((n0 * n0) + (n1 * n1)) + (n2 * n2);
                const n4 = s0.radius * s0.radius;
                if (n3 <= n4) {
                    return 0.0;
                }
                const n5 = ((n0 * r0.direction.v[0]) +
                    (n1 * r0.direction.v[1])) +
                    (n2 * r0.direction.v[2]);
                if (n5 < 0.0) {
                    return undefined;
                }
                const n6 = n3 - (n5 * n5);
                if (n6 > n4) {
                    return undefined;
                }
                const n7 = Math.sqrt(n4 - n6);
                return n5 - n7;
            }
            static intersectFrustum(ray, frustum) {
                throw Error('not implemented');
            }
            static create(position, direction) {
                return new Ray(position, direction);
            }
        }
        exports.Ray = Ray;
    });
    define("engine/math/box", ["require", "exports", "engine/math/vector3", "engine/math/ray"], function (require, exports, vector3_5, ray_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Box {
            constructor(min, max) {
                this.min = min;
                this.max = max;
            }
            toString() {
                return `[${this.min.toString()}, ${this.max.toString()}]`;
            }
            kind() {
                return 'Box';
            }
            clone() {
                return Box.clone(this);
            }
            corners() {
                return [
                    new vector3_5.Vector3(this.min.v[0], this.max.v[1], this.max.v[2]),
                    new vector3_5.Vector3(this.max.v[0], this.max.v[1], this.max.v[2]),
                    new vector3_5.Vector3(this.max.v[0], this.min.v[1], this.max.v[2]),
                    new vector3_5.Vector3(this.min.v[0], this.min.v[1], this.max.v[2]),
                    new vector3_5.Vector3(this.min.v[0], this.max.v[1], this.min.v[2]),
                    new vector3_5.Vector3(this.max.v[0], this.max.v[1], this.min.v[2]),
                    new vector3_5.Vector3(this.max.v[0], this.min.v[1], this.min.v[2]),
                    new vector3_5.Vector3(this.min.v[0], this.min.v[1], this.min.v[2])
                ];
            }
            static equals(b0, b1) {
                return (vector3_5.Vector3.equals(b0.min, b1.min) &&
                    vector3_5.Vector3.equals(b0.max, b1.max));
            }
            static merge(b0, b1) {
                return new Box(vector3_5.Vector3.min(b0.min, b1.min), vector3_5.Vector3.max(b0.max, b1.max));
            }
            static intersectBox(b0, b1) {
                if ((b0.max.v[0] < b1.min.v[0]) ||
                    (b0.min.v[0] > b1.max.v[0])) {
                    return false;
                }
                if ((b0.max.v[1] < b1.min.v[1]) ||
                    (b0.min.v[1] > b1.max.v[1])) {
                    return false;
                }
                return ((b0.max.v[2] >= b1.min.v[2]) &&
                    (b0.min.v[2] <= b1.max.v[2]));
            }
            static intersectPlane(b0, p0) {
                const n1 = new vector3_5.Vector3((p0.v[0] >= 0.0) ? b0.min.v[0] : b0.max.v[0], (p0.v[1] >= 0.0) ? b0.min.v[1] : b0.max.v[1], (p0.v[2] >= 0.0) ? b0.min.v[2] : b0.max.v[2]);
                const n0 = new vector3_5.Vector3((p0.v[0] >= 0.0) ? b0.max.v[0] : b0.min.v[0], (p0.v[1] >= 0.0) ? b0.max.v[1] : b0.min.v[1], (p0.v[2] >= 0.0) ? b0.max.v[2] : b0.min.v[2]);
                let n2 = ((p0.v[0] * n1.v[0]) +
                    (p0.v[1] * n1.v[1])) +
                    (p0.v[1] * n1.v[2]);
                if ((n2 + p0.v[3]) > 0.0) {
                    return 'front';
                }
                n2 = ((p0.v[0] * n0.v[0]) +
                    (p0.v[1] * n0.v[1])) +
                    (p0.v[2] * n0.v[2]);
                if ((n2 + p0.v[3]) < 0.0) {
                    return 'back';
                }
                return 'intersect';
            }
            static intersectRay(b0, r0) {
                return ray_2.Ray.intersectBox(r0, b0);
            }
            static intersectSphere(b0, s0) {
                const vector = vector3_5.Vector3.clamp(s0.position, b0.min, b0.max);
                const num = vector3_5.Vector3.distanceSq(s0.position, vector);
                return (num <= (s0.radius * s0.radius));
            }
            static fromSphere(s0) {
                return new Box(new vector3_5.Vector3(s0.position.v[0] - s0.radius, s0.position.v[1] - s0.radius, s0.position.v[2] - s0.radius), new vector3_5.Vector3(s0.position.v[0] + s0.radius, s0.position.v[1] + s0.radius, s0.position.v[2] + s0.radius));
            }
            static fromPoints(points) {
                let max = vector3_5.Vector3.MAX_VALUE;
                let min = vector3_5.Vector3.MIN_VALUE;
                for (let i = 0; i < points.length; i++) {
                    max = vector3_5.Vector3.min(max, points[i]);
                    min = vector3_5.Vector3.max(min, points[i]);
                }
                return new Box(min, max);
            }
            static clone(b0) {
                return new Box(b0.min.clone(), b0.max.clone());
            }
            static create(min, max) {
                return new Box(min, max);
            }
        }
        exports.Box = Box;
    });
    define("engine/math/plane", ["require", "exports", "engine/math/vector3", "engine/math/matrix"], function (require, exports, vector3_6, matrix_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Plane {
            constructor(a, b, c, d) {
                this.v = new Float32Array(4);
                this.v[0] = a;
                this.v[1] = b;
                this.v[2] = c;
                this.v[3] = d;
            }
            get a() {
                return this.v[0];
            }
            get b() {
                return this.v[1];
            }
            get c() {
                return this.v[2];
            }
            get d() {
                return this.v[3];
            }
            set a(value) {
                this.v[0] = value;
            }
            set b(value) {
                this.v[1] = value;
            }
            set c(value) {
                this.v[2] = value;
            }
            set d(value) {
                this.v[3] = value;
            }
            normal() {
                return new vector3_6.Vector3(this.v[0], this.v[1], this.v[2]);
            }
            toString() {
                return `[${this.v[0]}, ${this.v[1]}, ${this.v[2]}, ${this.v[3]}]`;
            }
            kind() {
                return 'Plane';
            }
            clone() {
                return Plane.clone(this);
            }
            static equals(p0, p1) {
                return ((p0.v[0] === p1.v[0]) &&
                    (p0.v[1] === p1.v[1]) &&
                    (p0.v[2] === p1.v[2]));
            }
            static normalize(p0) {
                const n0 = ((p0.v[0] * p0.v[0]) +
                    (p0.v[1] * p0.v[1])) +
                    (p0.v[2] * p0.v[2]);
                if (Math.abs((n0 - 1.0)) < 1.192093E-07) {
                    const p1 = Plane.zero();
                    p1.v[0] = p0.v[0];
                    p1.v[1] = p0.v[1];
                    p1.v[2] = p0.v[2];
                    p1.v[3] = p0.v[3];
                    return p1;
                }
                else {
                    const p1 = Plane.zero();
                    const n1 = 1.0 / Math.sqrt(n0);
                    p1.v[0] = p0.v[0] * n1;
                    p1.v[1] = p0.v[1] * n1;
                    p1.v[2] = p0.v[2] * n1;
                    p1.v[3] = p0.v[3] * n1;
                    return p1;
                }
            }
            static fromPoints(point1, point2, point3) {
                const p0 = Plane.zero();
                const n0 = point2.v[0] - point1.v[0];
                const n1 = point2.v[1] - point1.v[1];
                const n2 = point2.v[2] - point1.v[2];
                const n3 = point3.v[0] - point1.v[0];
                const n4 = point3.v[1] - point1.v[1];
                const n5 = point3.v[2] - point1.v[2];
                const n6 = (n1 * n5) - (n2 * n4);
                const n7 = (n2 * n3) - (n0 * n5);
                const n8 = (n0 * n4) - (n1 * n3);
                const n9 = ((n6 * n6) + (n7 * n7)) + (n8 * n8);
                const n10 = 1.0 / Math.sqrt(n9);
                p0.v[0] = n6 * n10;
                p0.v[1] = n7 * n10;
                p0.v[2] = n8 * n10;
                p0.v[3] = -(((p0.v[0] * point1.v[0]) +
                    (p0.v[1] * point1.v[1])) +
                    (p0.v[2] * point1.v[2]));
                return p0;
            }
            static transform(p0, m0) {
                const p1 = Plane.zero();
                const m1 = matrix_1.Matrix.invert(m0);
                const x = p0.v[0];
                const y = p0.v[1];
                const z = p0.v[2];
                const d = p0.v[3];
                p1.v[0] = (((x * m1.v[0]) + (y * m1.v[1])) + (z * m1.v[2])) + (d * m1.v[3]);
                p1.v[1] = (((x * m1.v[4]) + (y * m1.v[5])) + (z * m1.v[6])) + (d * m1.v[7]);
                p1.v[2] = (((x * m1.v[8]) + (y * m1.v[9])) + (z * m1.v[10])) + (d * m1.v[11]);
                p1.v[3] = (((x * m1.v[12]) + (y * m1.v[13])) + (z * m1.v[14])) + (d * m1.v[15]);
                return p1;
            }
            static dot4(p0, v0) {
                return ((((p0.v[0] * v0.v[0]) +
                    (p0.v[1] * v0.v[1])) +
                    (p0.v[2] * v0.v[2])) +
                    (p0.v[3] * v0.v[3]));
            }
            static dot3(p0, v0) {
                return ((((p0.v[0] * v0.v[0]) +
                    (p0.v[1] * v0.v[1])) +
                    (p0.v[2] * v0.v[2])) +
                    p0.v[3]);
            }
            static dotN(plane, normal) {
                return (((plane.v[0] * normal.v[0]) +
                    (plane.v[1] * normal.v[1])) +
                    (plane.v[2] * normal.v[2]));
            }
            static intersectBox(p0, b0) {
                const n0 = new vector3_6.Vector3((p0.v[0] >= 0.0) ? b0.min.v[0] : b0.max.v[0], (p0.v[1] >= 0.0) ? b0.min.v[1] : b0.max.v[1], (p0.v[2] >= 0.0) ? b0.min.v[2] : b0.max.v[2]);
                const n1 = new vector3_6.Vector3((p0.v[0] >= 0.0) ? b0.max.v[0] : b0.min.v[0], (p0.v[1] >= 0.0) ? b0.max.v[1] : b0.min.v[1], (p0.v[2] >= 0.0) ? b0.max.v[2] : b0.min.v[2]);
                let num = ((p0.v[0] * n0.v[0]) +
                    (p0.v[1] * n0.v[1])) +
                    (p0.v[2] * n0.v[2]);
                if ((num + p0.v[3]) > 0.0) {
                    return 'front';
                }
                num = ((p0.v[0] * n1.v[0]) +
                    (p0.v[1] * n1.v[1])) +
                    (p0.v[2] * n1.v[2]);
                if ((num + p0.v[3]) < 0.0) {
                    return 'back';
                }
                return 'intersect';
            }
            static intersectSphere(plane, sphere) {
                const n0 = ((sphere.position.v[0] * plane.v[0]) +
                    (sphere.position.v[1] * plane.v[1]) +
                    (sphere.position.v[2] * plane.v[2]));
                const n1 = n0 + plane.v[3];
                if (n1 > sphere.radius) {
                    return 'front';
                }
                if (n1 < -sphere.radius) {
                    return 'back';
                }
                return 'intersect';
            }
            static intersectFrustum(p0, f0) {
                let n0 = 0;
                for (let i = 0; i < 8; i++) {
                    const n1 = vector3_6.Vector3.dot(f0.corners[i], p0.normal());
                    if ((n1 + p0.v[3]) > 0.0) {
                        n0 |= 1;
                    }
                    else {
                        n0 |= 2;
                    }
                    if (n0 == 3) {
                        return 'intersect';
                    }
                }
                return (n0 == 1)
                    ? 'front'
                    : 'back';
            }
            static clone(p0) {
                return new Plane(p0.v[0], p0.v[1], p0.v[2], p0.v[3]);
            }
            static zero() {
                return Plane.create(0, 0, 0, 0);
            }
            static create(a, b, c, d) {
                return new Plane(a, b, c, d);
            }
        }
        exports.Plane = Plane;
    });
    define("engine/math/matrix", ["require", "exports", "engine/math/plane", "engine/math/vector3"], function (require, exports, plane_3, vector3_7) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Matrix {
            constructor(array) {
                this.v = new Float32Array(16);
                this.v.set(array, 0);
            }
            get m11() {
                return this.v[0];
            }
            get m12() {
                return this.v[1];
            }
            get m13() {
                return this.v[2];
            }
            get m14() {
                return this.v[3];
            }
            get m21() {
                return this.v[4];
            }
            get m22() {
                return this.v[5];
            }
            get m23() {
                return this.v[6];
            }
            get m24() {
                return this.v[7];
            }
            get m31() {
                return this.v[8];
            }
            get m32() {
                return this.v[9];
            }
            get m33() {
                return this.v[10];
            }
            get m34() {
                return this.v[11];
            }
            get m41() {
                return this.v[12];
            }
            get m42() {
                return this.v[13];
            }
            get m43() {
                return this.v[14];
            }
            get m44() {
                return this.v[15];
            }
            get up() {
                return new vector3_7.Vector3(this.v[4], this.v[5], this.v[6]);
            }
            get down() {
                return new vector3_7.Vector3(-this.v[4], -this.v[5], -this.v[6]);
            }
            get right() {
                return new vector3_7.Vector3(this.v[0], this.v[1], this.v[2]);
            }
            get left() {
                return new vector3_7.Vector3(-this.v[0], -this.v[1], -this.v[2]);
            }
            get forward() {
                return new vector3_7.Vector3(-this.v[8], -this.v[9], -this.v[10]);
            }
            get back() {
                return new vector3_7.Vector3(this.v[8], this.v[9], this.v[10]);
            }
            get translation() {
                return new vector3_7.Vector3(this.v[12], this.v[13], this.v[14]);
            }
            set m11(value) {
                this.v[0] = value;
            }
            set m12(value) {
                this.v[1] = value;
            }
            set m13(value) {
                this.v[2] = value;
            }
            set m14(value) {
                this.v[3] = value;
            }
            set m21(value) {
                this.v[4] = value;
            }
            set m22(value) {
                this.v[5] = value;
            }
            set m23(value) {
                this.v[6] = value;
            }
            set m24(value) {
                this.v[7] = value;
            }
            set m31(value) {
                this.v[8] = value;
            }
            set m32(value) {
                this.v[9] = value;
            }
            set m33(value) {
                this.v[10] = value;
            }
            set m34(value) {
                this.v[11] = value;
            }
            set m41(value) {
                this.v[12] = value;
            }
            set m42(value) {
                this.v[13] = value;
            }
            set m43(value) {
                this.v[14] = value;
            }
            set m44(value) {
                this.v[15] = value;
            }
            translate(v0) {
                return Matrix.mul(this, Matrix.translation(v0));
            }
            scale(v0) {
                return Matrix.mul(this, Matrix.scale(v0));
            }
            rotateX(radian) {
                return Matrix.mul(this, Matrix.rotationX(radian));
            }
            rotateY(radian) {
                return Matrix.mul(this, Matrix.rotationY(radian));
            }
            rotateZ(radian) {
                return Matrix.mul(this, Matrix.rotationZ(radian));
            }
            toString() {
                const buffer = [];
                buffer.push(`[${this.v[0]}, ${this.v[1]}, ${this.v[2]}, ${this.v[3]}`);
                buffer.push(` ${this.v[4]}, ${this.v[5]}, ${this.v[6]}, ${this.v[7]}`);
                buffer.push(` ${this.v[8]}, ${this.v[9]}, ${this.v[10]}, ${this.v[11]}`);
                buffer.push(` ${this.v[12]}, ${this.v[13]}, ${this.v[14]}, ${this.v[15]}]`);
                return buffer.join('\n');
            }
            kind() {
                return 'Matrix';
            }
            clone() {
                return Matrix.clone(this);
            }
            static identity() {
                return new Matrix([
                    1, 0, 0, 0,
                    0, 1, 0, 0,
                    0, 0, 1, 0,
                    0, 0, 0, 1
                ]);
            }
            static translation(v0) {
                return new Matrix([
                    1, 0, 0, 0,
                    0, 1, 0, 0,
                    0, 0, 1, 0,
                    v0.v[0], v0.v[1], v0.v[2], 1
                ]);
            }
            static scale(v0) {
                return new Matrix([
                    v0.v[0], 0, 0, 0,
                    0, v0.v[1], 0, 0,
                    0, 0, v0.v[2], 0,
                    0, 0, 0, 1
                ]);
            }
            static rotationX(radians) {
                const cos = Math.cos(radians);
                const sin = Math.sin(radians);
                return new Matrix([
                    1, 0, 0, 0,
                    0, cos, sin, 0,
                    0, -sin, cos, 0,
                    0, 0, 0, 1
                ]);
            }
            static rotationY(radians) {
                const cos = Math.cos(radians);
                const sin = Math.sin(radians);
                return new Matrix([
                    cos, 0, -sin, 0,
                    0, 1, 0, 0,
                    sin, 0, cos, 0,
                    0, 0, 0, 1
                ]);
            }
            static rotationZ(radians) {
                const cos = Math.cos(radians);
                const sin = Math.sin(radians);
                return new Matrix([
                    cos, sin, 0, 0,
                    -sin, cos, 0, 0,
                    0, 0, 1, 0,
                    0, 0, 0, 1
                ]);
            }
            static fromAxisAngle(axis, radians) {
                const x = axis.v[0];
                const y = axis.v[1];
                const z = axis.v[2];
                const n0 = Math.sin(radians);
                const n1 = Math.cos(radians);
                const n2 = x * x;
                const n3 = y * y;
                const n4 = z * z;
                const n5 = x * y;
                const n6 = x * z;
                const n7 = y * z;
                return new Matrix([
                    n2 + (n1 * (1.0 - n2)),
                    (n5 - (n1 * n5)) + (n0 * z),
                    (n6 - (n1 * n6)) - (n0 * y),
                    0.0,
                    (n5 - (n1 * n5)) - (n0 * z),
                    n3 + (n1 * (1.0 - n3)),
                    (n7 - (n1 * n7)) + (n0 * x),
                    0.0,
                    (n6 - (n1 * n6)) + (n0 * y),
                    (n7 - (n1 * n7)) - (n0 * x),
                    n4 + (n1 * (1.0 - n4)),
                    0.0,
                    0.0, 0.0, 0.0, 1.0
                ]);
            }
            static perspectiveFov(fov, aspect, near, far) {
                const n0 = 1.0 / Math.tan(fov * 0.5);
                const n1 = n0 / aspect;
                const m0 = Matrix.identity();
                m0.v[0] = n1;
                m0.v[1] = 0;
                m0.v[2] = 0;
                m0.v[3] = 0;
                m0.v[5] = n0;
                m0.v[4] = 0;
                m0.v[6] = 0;
                m0.v[7] = 0;
                m0.v[8] = 0;
                m0.v[9] = 0;
                m0.v[10] = far / (near - far);
                m0.v[11] = -1;
                m0.v[12] = 0;
                m0.v[13] = 0;
                m0.v[15] = 0;
                m0.v[14] = (near * far) / (near - far);
                return m0;
            }
            static perspective(width, height, near, far) {
                const m0 = Matrix.identity();
                m0.v[0] = (2.0 * near) / width;
                m0.v[1] = m0.v[2] = m0.v[3] = 0.0;
                m0.v[5] = (2.0 * near) / height;
                m0.v[4] = m0.v[6] = m0.v[7] = 0.0;
                m0.v[10] = far / (near - far);
                m0.v[8] = m0.v[9] = 0.0;
                m0.v[11] = -1.0;
                m0.v[12] = m0.v[13] = m0.v[15] = 0.0;
                m0.v[14] = (near * far) / (near - far);
                return m0;
            }
            static perspectiveOffset(left, right, bottom, top, near, far) {
                const m0 = Matrix.identity();
                m0.v[0] = (2.0 * near) / (right - left);
                m0.v[1] = m0.v[2] = m0.v[3] = 0.0;
                m0.v[5] = (2.0 * near) / (top - bottom);
                m0.v[4] = m0.v[6] = m0.v[7] = 0.0;
                m0.v[8] = (left + right) / (right - left);
                m0.v[9] = (top + bottom) / (top - bottom);
                m0.v[10] = far / (near - far);
                m0.v[11] = -1.0;
                m0.v[14] = (near * far) / (near - far);
                m0.v[12] = m0.v[13] = m0.v[15] = 0.0;
                return m0;
            }
            static orthographic(width, height, near, far) {
                const m0 = Matrix.identity();
                m0.v[0] = 2.0 / width;
                m0.v[1] = m0.v[2] = m0.v[3] = 0.0;
                m0.v[5] = 2.0 / height;
                m0.v[4] = m0.v[6] = m0.v[7] = 0.0;
                m0.v[10] = 1.0 / (near - far);
                m0.v[8] = m0.v[9] = m0.v[11] = 0.0;
                m0.v[12] = m0.v[13] = 0.0;
                m0.v[14] = near / (near - far);
                m0.v[15] = 1.0;
                return m0;
            }
            static orthographicOffset(left, right, bottom, top, near, far) {
                const m0 = Matrix.identity();
                m0.v[0] = 2.0 / (right - left);
                m0.v[1] = m0.v[2] = m0.v[3] = 0.0;
                m0.v[5] = 2.0 / (top - bottom);
                m0.v[4] = m0.v[6] = m0.v[7] = 0.0;
                m0.v[10] = 1.0 / (near - far);
                m0.v[8] = m0.v[9] = m0.v[11] = 0.0;
                m0.v[12] = (left + right) / (left - right);
                m0.v[13] = (top + bottom) / (bottom - top);
                m0.v[14] = near / (near - far);
                m0.v[15] = 1.0;
                return m0;
            }
            static lookAt(position, target, up) {
                const m0 = Matrix.identity();
                const v0 = vector3_7.Vector3.normalize(vector3_7.Vector3.sub(position, target));
                const v1 = vector3_7.Vector3.normalize(vector3_7.Vector3.cross(up, v0));
                const v2 = vector3_7.Vector3.cross(v0, v1);
                m0.v[0] = v1.v[0];
                m0.v[1] = v2.v[0];
                m0.v[2] = v0.v[0];
                m0.v[3] = 0.0;
                m0.v[4] = v1.v[1];
                m0.v[5] = v2.v[1];
                m0.v[6] = v0.v[1];
                m0.v[7] = 0.0;
                m0.v[8] = v1.v[2];
                m0.v[9] = v2.v[2];
                m0.v[10] = v0.v[2];
                m0.v[11] = 0.0;
                m0.v[12] = -vector3_7.Vector3.dot(v1, position);
                m0.v[13] = -vector3_7.Vector3.dot(v2, position);
                m0.v[14] = -vector3_7.Vector3.dot(v0, position);
                m0.v[15] = 1.0;
                return m0;
            }
            static world(position, forward, up) {
                const m0 = Matrix.identity();
                const v0 = vector3_7.Vector3.normalize(vector3_7.Vector3.sub(position, forward));
                const v1 = vector3_7.Vector3.normalize(vector3_7.Vector3.cross(up, v0));
                const v2 = vector3_7.Vector3.cross(v0, v1);
                m0.v[0] = v1.v[0];
                m0.v[1] = v1.v[1];
                m0.v[2] = v1.v[2];
                m0.v[4] = v2.v[0];
                m0.v[5] = v2.v[1];
                m0.v[6] = v2.v[2];
                m0.v[8] = v0.v[0];
                m0.v[9] = v0.v[1];
                m0.v[10] = v0.v[2];
                m0.v[12] = 0.0;
                m0.v[13] = 0.0;
                m0.v[14] = 0.0;
                m0.v[15] = 1.0;
                m0.v[3] = -vector3_7.Vector3.dot(v1, position);
                m0.v[7] = -vector3_7.Vector3.dot(v2, position);
                m0.v[11] = -vector3_7.Vector3.dot(v0, position);
                return m0;
            }
            static fromQuaternion(q0) {
                const m0 = Matrix.identity();
                const n0 = q0.v[0] * q0.v[0];
                const n1 = q0.v[1] * q0.v[1];
                const n2 = q0.v[2] * q0.v[2];
                const n3 = q0.v[0] * q0.v[1];
                const n4 = q0.v[2] * q0.v[3];
                const n5 = q0.v[2] * q0.v[0];
                const n6 = q0.v[1] * q0.v[3];
                const n7 = q0.v[1] * q0.v[2];
                const n8 = q0.v[0] * q0.v[3];
                m0.v[0] = 1.0 - (2.0 * (n1 + n2));
                m0.v[1] = 2.0 * (n3 + n4);
                m0.v[2] = 2.0 * (n5 - n6);
                m0.v[3] = 0.0;
                m0.v[4] = 2.0 * (n3 - n4);
                m0.v[5] = 1.0 - (2.0 * (n2 + n0));
                m0.v[6] = 2.0 * (n7 + n8);
                m0.v[7] = 0.0;
                m0.v[8] = 2.0 * (n5 + n6);
                m0.v[9] = 2.0 * (n7 - n8);
                m0.v[10] = 1.0 - (2.0 * (n1 + n0));
                m0.v[11] = 0.0;
                m0.v[12] = 0.0;
                m0.v[13] = 0.0;
                m0.v[14] = 0.0;
                m0.v[15] = 1.0;
                return m0;
            }
            static reflection(p0) {
                const m0 = Matrix.identity();
                const p1 = plane_3.Plane.normalize(p0);
                const x = p1.v[0];
                const y = p1.v[1];
                const z = p1.v[2];
                const n0 = -2.0 * x;
                const n1 = -2.0 * y;
                const n2 = -2.0 * z;
                m0.v[0] = (n0 * x) + 1.0;
                m0.v[1] = n1 * x;
                m0.v[2] = n2 * x;
                m0.v[3] = 0.0;
                m0.v[4] = n0 * y;
                m0.v[5] = (n1 * y) + 1.0;
                m0.v[6] = n2 * y;
                m0.v[7] = 0.0;
                m0.v[8] = n0 * z;
                m0.v[9] = n1 * z;
                m0.v[10] = (n2 * z) + 1.0;
                m0.v[11] = 0.0;
                m0.v[12] = n0 * p1.v[3];
                m0.v[13] = n1 * p1.v[3];
                m0.v[14] = n2 * p1.v[3];
                m0.v[15] = 1.0;
                return m0;
            }
            static invert(m0) {
                const m1 = Matrix.identity();
                const n0 = m0.v[0];
                const n1 = m0.v[1];
                const n2 = m0.v[2];
                const n3 = m0.v[3];
                const n4 = m0.v[4];
                const n5 = m0.v[5];
                const n6 = m0.v[6];
                const n7 = m0.v[7];
                const n8 = m0.v[8];
                const n9 = m0.v[9];
                const n10 = m0.v[10];
                const n11 = m0.v[11];
                const n12 = m0.v[12];
                const n13 = m0.v[13];
                const n14 = m0.v[14];
                const n15 = m0.v[15];
                const n16 = (n10 * n15) - (n11 * n14);
                const n17 = (n9 * n15) - (n11 * n13);
                const n18 = (n9 * n14) - (n10 * n13);
                const n19 = (n8 * n15) - (n11 * n12);
                const n20 = (n8 * n14) - (n10 * n12);
                const n21 = (n8 * n13) - (n9 * n12);
                const n22 = ((n5 * n16) - (n6 * n17)) + (n7 * n18);
                const n23 = -(((n4 * n16) - (n6 * n19)) + (n7 * n20));
                const n24 = ((n4 * n17) - (n5 * n19)) + (n7 * n21);
                const n25 = -(((n4 * n18) - (n5 * n20)) + (n6 * n21));
                const n26 = 1.0 / ((((n0 * n22) + (n1 * n23)) + (n2 * n24)) + (n3 * n25));
                m1.v[0] = n22 * n26;
                m1.v[4] = n23 * n26;
                m1.v[8] = n24 * n26;
                m1.v[12] = n25 * n26;
                m1.v[1] = -(((n1 * n16) - (n2 * n17)) + (n3 * n18)) * n26;
                m1.v[5] = (((n0 * n16) - (n2 * n19)) + (n3 * n20)) * n26;
                m1.v[9] = -(((n0 * n17) - (n1 * n19)) + (n3 * n21)) * n26;
                m1.v[13] = (((n0 * n18) - (n1 * n20)) + (n2 * n21)) * n26;
                const n27 = (n6 * n15) - (n7 * n14);
                const n28 = (n5 * n15) - (n7 * n13);
                const n29 = (n5 * n14) - (n6 * n13);
                const n30 = (n4 * n15) - (n7 * n12);
                const n32 = (n4 * n14) - (n6 * n12);
                const n33 = (n4 * n13) - (n5 * n12);
                m1.v[2] = (((n1 * n27) - (n2 * n28)) + (n3 * n29)) * n26;
                m1.v[6] = -(((n0 * n27) - (n2 * n30)) + (n3 * n32)) * n26;
                m1.v[10] = (((n0 * n28) - (n1 * n30)) + (n3 * n33)) * n26;
                m1.v[14] = -(((n0 * n29) - (n1 * n32)) + (n2 * n33)) * n26;
                const n34 = (n6 * n11) - (n7 * n10);
                const n35 = (n5 * n11) - (n7 * n9);
                const n36 = (n5 * n10) - (n6 * n9);
                const n37 = (n4 * n11) - (n7 * n8);
                const n38 = (n4 * n10) - (n6 * n8);
                const n39 = (n4 * n9) - (n5 * n8);
                m1.v[3] = -(((n1 * n34) - (n2 * n35)) + (n3 * n36)) * n26;
                m1.v[7] = (((n0 * n34) - (n2 * n37)) + (n3 * n38)) * n26;
                m1.v[11] = -(((n0 * n35) - (n1 * n37)) + (n3 * n39)) * n26;
                m1.v[15] = (((n0 * n36) - (n1 * n38)) + (n2 * n39)) * n26;
                return m1;
            }
            static transpose(m0) {
                const m1 = Matrix.identity();
                m1.v[0] = m0.v[0];
                m1.v[1] = m0.v[4];
                m1.v[2] = m0.v[8];
                m1.v[3] = m0.v[12];
                m1.v[4] = m0.v[1];
                m1.v[5] = m0.v[5];
                m1.v[6] = m0.v[9];
                m1.v[7] = m0.v[13];
                m1.v[8] = m0.v[2];
                m1.v[9] = m0.v[6];
                m1.v[10] = m0.v[10];
                m1.v[11] = m0.v[14];
                m1.v[12] = m0.v[3];
                m1.v[13] = m0.v[7];
                m1.v[14] = m0.v[11];
                m1.v[15] = m0.v[15];
                return m1;
            }
            static determinant(m0) {
                const n0 = m0.v[0];
                const n1 = m0.v[1];
                const n2 = m0.v[2];
                const n3 = m0.v[3];
                const n4 = m0.v[4];
                const n5 = m0.v[5];
                const n6 = m0.v[6];
                const n7 = m0.v[7];
                const n8 = m0.v[8];
                const n9 = m0.v[9];
                const n10 = m0.v[10];
                const n11 = m0.v[11];
                const n12 = m0.v[12];
                const n13 = m0.v[13];
                const n14 = m0.v[14];
                const n15 = m0.v[15];
                const n16 = (n10 * n15) - (n11 * n14);
                const n17 = (n9 * n15) - (n11 * n13);
                const n18 = (n9 * n14) - (n10 * n13);
                const n19 = (n8 * n15) - (n11 * n12);
                const n20 = (n8 * n14) - (n10 * n12);
                const n21 = (n8 * n13) - (n9 * n12);
                return ((((n0 * (((n5 * n16) - (n6 * n17)) + (n7 * n18))) -
                    (n1 * (((n4 * n16) - (n6 * n19)) + (n7 * n20)))) + (n2 * (((n4 * n17) -
                    (n5 * n19)) + (n7 * n21)))) - (n3 * (((n4 * n18) - (n5 * n20)) + (n6 * n21))));
            }
            static lerp(m0, m1, amount) {
                const m2 = Matrix.identity();
                m2.v[0] = m0.v[0] + ((m1.v[0] - m0.v[0]) * amount);
                m2.v[1] = m0.v[1] + ((m1.v[1] - m0.v[1]) * amount);
                m2.v[2] = m0.v[2] + ((m1.v[2] - m0.v[2]) * amount);
                m2.v[3] = m0.v[3] + ((m1.v[3] - m0.v[3]) * amount);
                m2.v[4] = m0.v[4] + ((m1.v[4] - m0.v[4]) * amount);
                m2.v[5] = m0.v[5] + ((m1.v[5] - m0.v[5]) * amount);
                m2.v[6] = m0.v[6] + ((m1.v[6] - m0.v[6]) * amount);
                m2.v[7] = m0.v[7] + ((m1.v[7] - m0.v[7]) * amount);
                m2.v[8] = m0.v[8] + ((m1.v[8] - m0.v[8]) * amount);
                m2.v[9] = m0.v[9] + ((m1.v[9] - m0.v[9]) * amount);
                m2.v[10] = m0.v[10] + ((m1.v[10] - m0.v[10]) * amount);
                m2.v[11] = m0.v[11] + ((m1.v[11] - m0.v[11]) * amount);
                m2.v[12] = m0.v[12] + ((m1.v[12] - m0.v[12]) * amount);
                m2.v[13] = m0.v[13] + ((m1.v[13] - m0.v[13]) * amount);
                m2.v[14] = m0.v[14] + ((m1.v[14] - m0.v[14]) * amount);
                m2.v[15] = m0.v[15] + ((m1.v[15] - m0.v[15]) * amount);
                return m2;
            }
            static negate(m0) {
                const m1 = Matrix.identity();
                m1.v[0] = -m0.v[0];
                m1.v[1] = -m0.v[1];
                m1.v[2] = -m0.v[2];
                m1.v[3] = -m0.v[3];
                m1.v[4] = -m0.v[4];
                m1.v[5] = -m0.v[5];
                m1.v[6] = -m0.v[6];
                m1.v[7] = -m0.v[7];
                m1.v[8] = -m0.v[8];
                m1.v[9] = -m0.v[9];
                m1.v[10] = -m0.v[10];
                m1.v[11] = -m0.v[11];
                m1.v[12] = -m0.v[12];
                m1.v[13] = -m0.v[13];
                m1.v[14] = -m0.v[14];
                m1.v[15] = -m0.v[15];
                return m1;
            }
            static fast_add(m0, m1, out) {
                out.v[0] = m0.v[0] + m1.v[0];
                out.v[1] = m0.v[1] + m1.v[1];
                out.v[2] = m0.v[2] + m1.v[2];
                out.v[3] = m0.v[3] + m1.v[3];
                out.v[4] = m0.v[4] + m1.v[4];
                out.v[5] = m0.v[5] + m1.v[5];
                out.v[6] = m0.v[6] + m1.v[6];
                out.v[7] = m0.v[7] + m1.v[7];
                out.v[8] = m0.v[8] + m1.v[8];
                out.v[9] = m0.v[9] + m1.v[9];
                out.v[10] = m0.v[10] + m1.v[10];
                out.v[11] = m0.v[11] + m1.v[11];
                out.v[12] = m0.v[12] + m1.v[12];
                out.v[13] = m0.v[13] + m1.v[13];
                out.v[14] = m0.v[14] + m1.v[14];
                out.v[15] = m0.v[15] + m1.v[15];
            }
            static add(m0, m1) {
                const out = Matrix.identity();
                Matrix.fast_add(m0, m1, out);
                return out;
            }
            static fast_sub(m0, m1, out) {
                out.v[0] = m0.v[0] - m1.v[0];
                out.v[1] = m0.v[1] - m1.v[1];
                out.v[2] = m0.v[2] - m1.v[2];
                out.v[3] = m0.v[3] - m1.v[3];
                out.v[4] = m0.v[4] - m1.v[4];
                out.v[5] = m0.v[5] - m1.v[5];
                out.v[6] = m0.v[6] - m1.v[6];
                out.v[7] = m0.v[7] - m1.v[7];
                out.v[8] = m0.v[8] - m1.v[8];
                out.v[9] = m0.v[9] - m1.v[9];
                out.v[10] = m0.v[10] - m1.v[10];
                out.v[11] = m0.v[11] - m1.v[11];
                out.v[12] = m0.v[12] - m1.v[12];
                out.v[13] = m0.v[13] - m1.v[13];
                out.v[14] = m0.v[14] - m1.v[14];
                out.v[15] = m0.v[15] - m1.v[15];
            }
            static sub(m0, m1) {
                const out = Matrix.identity();
                Matrix.fast_sub(m0, m1, out);
                return out;
            }
            static fast_mul(m0, m1, out) {
                out.v[0] = (((m0.v[0] * m1.v[0]) + (m0.v[1] * m1.v[4])) + (m0.v[2] * m1.v[8])) + (m0.v[3] * m1.v[12]);
                out.v[1] = (((m0.v[0] * m1.v[1]) + (m0.v[1] * m1.v[5])) + (m0.v[2] * m1.v[9])) + (m0.v[3] * m1.v[13]);
                out.v[2] = (((m0.v[0] * m1.v[2]) + (m0.v[1] * m1.v[6])) + (m0.v[2] * m1.v[10])) + (m0.v[3] * m1.v[14]);
                out.v[3] = (((m0.v[0] * m1.v[3]) + (m0.v[1] * m1.v[7])) + (m0.v[2] * m1.v[11])) + (m0.v[3] * m1.v[15]);
                out.v[4] = (((m0.v[4] * m1.v[0]) + (m0.v[5] * m1.v[4])) + (m0.v[6] * m1.v[8])) + (m0.v[7] * m1.v[12]);
                out.v[5] = (((m0.v[4] * m1.v[1]) + (m0.v[5] * m1.v[5])) + (m0.v[6] * m1.v[9])) + (m0.v[7] * m1.v[13]);
                out.v[6] = (((m0.v[4] * m1.v[2]) + (m0.v[5] * m1.v[6])) + (m0.v[6] * m1.v[10])) + (m0.v[7] * m1.v[14]);
                out.v[7] = (((m0.v[4] * m1.v[3]) + (m0.v[5] * m1.v[7])) + (m0.v[6] * m1.v[11])) + (m0.v[7] * m1.v[15]);
                out.v[8] = (((m0.v[8] * m1.v[0]) + (m0.v[9] * m1.v[4])) + (m0.v[10] * m1.v[8])) + (m0.v[11] * m1.v[12]);
                out.v[9] = (((m0.v[8] * m1.v[1]) + (m0.v[9] * m1.v[5])) + (m0.v[10] * m1.v[9])) + (m0.v[11] * m1.v[13]);
                out.v[10] = (((m0.v[8] * m1.v[2]) + (m0.v[9] * m1.v[6])) + (m0.v[10] * m1.v[10])) + (m0.v[11] * m1.v[14]);
                out.v[11] = (((m0.v[8] * m1.v[3]) + (m0.v[9] * m1.v[7])) + (m0.v[10] * m1.v[11])) + (m0.v[11] * m1.v[15]);
                out.v[12] = (((m0.v[12] * m1.v[0]) + (m0.v[13] * m1.v[4])) + (m0.v[14] * m1.v[8])) + (m0.v[15] * m1.v[12]);
                out.v[13] = (((m0.v[12] * m1.v[1]) + (m0.v[13] * m1.v[5])) + (m0.v[14] * m1.v[9])) + (m0.v[15] * m1.v[13]);
                out.v[14] = (((m0.v[12] * m1.v[2]) + (m0.v[13] * m1.v[6])) + (m0.v[14] * m1.v[10])) + (m0.v[15] * m1.v[14]);
                out.v[15] = (((m0.v[12] * m1.v[3]) + (m0.v[13] * m1.v[7])) + (m0.v[14] * m1.v[11])) + (m0.v[15] * m1.v[15]);
            }
            static mul(m0, m1) {
                const out = Matrix.identity();
                Matrix.fast_mul(m0, m1, out);
                return out;
            }
            static fast_div(m0, m1, out) {
                out.v[0] = m0.v[0] / m1.v[0];
                out.v[1] = m0.v[1] / m1.v[1];
                out.v[2] = m0.v[2] / m1.v[2];
                out.v[3] = m0.v[3] / m1.v[3];
                out.v[4] = m0.v[4] / m1.v[4];
                out.v[5] = m0.v[5] / m1.v[5];
                out.v[6] = m0.v[6] / m1.v[6];
                out.v[7] = m0.v[7] / m1.v[7];
                out.v[8] = m0.v[8] / m1.v[8];
                out.v[9] = m0.v[9] / m1.v[9];
                out.v[10] = m0.v[10] / m1.v[10];
                out.v[11] = m0.v[11] / m1.v[11];
                out.v[12] = m0.v[12] / m1.v[12];
                out.v[13] = m0.v[13] / m1.v[13];
                out.v[14] = m0.v[14] / m1.v[14];
                out.v[15] = m0.v[15] / m1.v[15];
            }
            static div(m0, m1) {
                const out = Matrix.identity();
                Matrix.fast_div(m0, m1, out);
                return out;
            }
            static fast_clone(m0, out) {
                for (let i = 0; i < 16; i += 1) {
                    out.v[i] = m0.v[i];
                }
            }
            static clone(m0) {
                return new Matrix([
                    m0.v[0], m0.v[1], m0.v[2], m0.v[3],
                    m0.v[4], m0.v[5], m0.v[6], m0.v[7],
                    m0.v[8], m0.v[9], m0.v[10], m0.v[11],
                    m0.v[12], m0.v[13], m0.v[14], m0.v[15]
                ]);
            }
            static create(array) {
                return new Matrix(array);
            }
        }
        exports.Matrix = Matrix;
    });
    define("engine/math/vector2", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const f32 = { min: Number.MIN_VALUE, max: Number.MAX_VALUE };
        class Vector2 {
            constructor(x, y) {
                this.v = new Float32Array(2);
                this.v[0] = x;
                this.v[1] = y;
            }
            toString() {
                return `[${this.v[0]}, ${this.v[1]}]`;
            }
            kind() {
                return 'Vector2';
            }
            clone() {
                return Vector2.clone(this);
            }
            get x() {
                return this.v[0];
            }
            get y() {
                return this.v[1];
            }
            set x(value) {
                this.v[0] = value;
            }
            set y(value) {
                this.v[1] = value;
            }
            length() {
                return Vector2.getLength(this);
            }
            lengthSq() {
                return Vector2.getLengthSq(this);
            }
            normalize() {
                return Vector2.normalize(this);
            }
            dot(v0) {
                return Vector2.dot(this, v0);
            }
            add(v0) {
                return Vector2.add(this, v0);
            }
            sub(v0) {
                return Vector2.sub(this, v0);
            }
            mul(v0) {
                return Vector2.mul(this, v0);
            }
            div(v0) {
                return Vector2.div(this, v0);
            }
            scale(s0) {
                return Vector2.scale(this, s0);
            }
            negate() {
                return Vector2.negate(this);
            }
            static zero() {
                return new Vector2(0.0, 0.0);
            }
            static one() {
                return new Vector2(1.0, 1.0);
            }
            static unitX() {
                return new Vector2(1.0, 0.0);
            }
            static unitY() {
                return new Vector2(0.0, 1.0);
            }
            static left() {
                return new Vector2(-1.0, 0.0);
            }
            static right() {
                return new Vector2(1.0, 0.0);
            }
            static up() {
                return new Vector2(0.0, 1.0);
            }
            static down() {
                return new Vector2(0.0, -1.0);
            }
            static equals(v0, v1) {
                return (v0.v[0] === v1.v[0] &&
                    v0.v[1] === v1.v[1]);
            }
            static getLength(v0) {
                return Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]));
            }
            static getLengthSq(v0) {
                return ((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]));
            }
            static distance(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                return Math.sqrt((x * x) + (y * y));
            }
            static distanceSq(v0, v1) {
                const x = v0.v[0] - v1.v[0];
                const y = v0.v[1] - v1.v[1];
                return ((x * x) + (y * y));
            }
            static dot(v0, v1) {
                return ((v0.v[0] * v1.v[0]) +
                    (v0.v[1] * v1.v[1]));
            }
            static normalize(v0) {
                const len = 1.0 / Math.sqrt((v0.v[0] * v0.v[0]) +
                    (v0.v[1] * v0.v[1]));
                return new Vector2(v0.v[0] * len, v0.v[1] * len);
            }
            static reflect(v0, n0) {
                const dot = ((v0.v[0] * n0.v[0]) +
                    (v0.v[1] * n0.v[1]));
                return new Vector2(v0.v[0] - ((2.0 * dot) * n0.v[0]), v0.v[1] - ((2.0 * dot) * n0.v[1]));
            }
            static abs(v0) {
                return new Vector2(Math.abs(v0.v[0]), Math.abs(v0.v[1]));
            }
            static min(v0, v1) {
                return new Vector2((v0.v[0] < v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] < v1.v[1]) ? v0.v[1] : v1.v[1]);
            }
            static max(v0, v1) {
                return new Vector2((v0.v[0] > v1.v[0]) ? v0.v[0] : v1.v[0], (v0.v[1] > v1.v[1]) ? v0.v[1] : v1.v[1]);
            }
            static clamp(v0, min, max) {
                let x = v0.v[0];
                let y = v0.v[1];
                x = (x > max.v[0]) ? max.v[0] : x;
                x = (x < min.v[0]) ? min.v[0] : x;
                y = (y > max.v[1]) ? max.v[1] : y;
                y = (y < min.v[1]) ? min.v[1] : y;
                return new Vector2(x, y);
            }
            static lerp(v0, v1, amount) {
                return new Vector2(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount));
            }
            static barycentric(v0, v1, v2, amount0, amount1) {
                return new Vector2((v0.v[0] + (amount0 * (v1.v[0] - v0.v[0]))) + (amount1 * (v2.v[0] - v0.v[0])), (v0.v[1] + (amount0 * (v1.v[1] - v0.v[1]))) + (amount1 * (v2.v[1] - v0.v[1])));
            }
            static smoothstep(v0, v1, amount) {
                amount = (amount > 1.0) ? 1.0 : ((amount < 0.0) ? 0.0 : amount);
                amount = (amount * amount) * (3.0 - (2.0 * amount));
                return new Vector2(v0.v[0] + ((v1.v[0] - v0.v[0]) * amount), v0.v[1] + ((v1.v[1] - v0.v[1]) * amount));
            }
            static catmullrom(v0, v1, v2, v3, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                return new Vector2(0.5 * ((((2.0 * v1.v[0])
                    + ((-v0.v[0] + v2.v[0]) * amount))
                    + (((((2.0 * v0.v[0]) - (5.0 * v1.v[0]))
                        + (4.0 * v2.v[0])) - v3.v[0]) * n0))
                    + ((((-v0.v[0] + (3.0 * v1.v[0]))
                        - (3.0 * v2.v[0])) + v3.v[0]) * n1)), 0.5 * ((((2.0 * v1.v[1])
                    + ((-v0.v[1] + v2.v[1]) * amount))
                    + (((((2.0 * v0.v[1]) - (5.0 * v1.v[1]))
                        + (4.0 * v2.v[1])) - v3.v[1]) * n0))
                    + ((((-v0.v[1] + (3.0 * v1.v[1]))
                        - (3.0 * v2.v[1])) + v3.v[1]) * n1)));
            }
            static hermite(v0, t0, v1, t1, amount) {
                const n0 = amount * amount;
                const n1 = amount * n0;
                const n2 = ((2.0 * n1) - (3.0 * n0)) + 1.0;
                const n3 = (-2.0 * n1) + (3.0 * n0);
                const n4 = (n1 - (2.0 * n0)) + amount;
                const n5 = n1 - n0;
                return new Vector2((((v0.v[0] * n2) + (v1.v[0] * n3)) + (t0.v[0] * n4)) + (t1.v[0] * n5), (((v0.v[1] * n2) + (v1.v[1] * n3)) + (t0.v[1] * n4)) + (t1.v[1] * n5));
            }
            static transform(v0, m0) {
                return new Vector2(((v0.v[0] * m0.v[0]) + (v0.v[1] * m0.v[4])) + m0.v[12], ((v0.v[0] * m0.v[1]) + (v0.v[1] * m0.v[5])) + m0.v[13]);
            }
            static transformNormal(n0, m0) {
                return new Vector2((n0.v[0] * m0.v[0]) + (n0.v[1] * m0.v[4]), (n0.v[0] * m0.v[1]) + (n0.v[1] * m0.v[5]));
            }
            static transformQuaternion(v0, q0) {
                const n0 = q0.v[0] + q0.v[0];
                const n1 = q0.v[1] + q0.v[1];
                const n2 = q0.v[2] + q0.v[2];
                const n3 = q0.v[3] * n2;
                const n4 = q0.v[0] * n0;
                const n5 = q0.v[0] * n1;
                const n6 = q0.v[1] * n1;
                const n7 = q0.v[2] * n2;
                return new Vector2((v0.v[0] * ((1.0 - n6) - n7)) + (v0.v[1] * (n5 - n3)), (v0.v[0] * (n5 + n3)) + (v0.v[1] * ((1.0 - n4) - n7)));
            }
            static add(v0, v1) {
                return new Vector2(v0.v[0] + v1.v[0], v0.v[1] + v1.v[1]);
            }
            static sub(v0, v1) {
                return new Vector2(v0.v[0] - v1.v[0], v0.v[1] - v1.v[1]);
            }
            static mul(v0, v1) {
                return new Vector2(v0.v[0] * v1.v[0], v0.v[1] * v1.v[1]);
            }
            static div(v0, v1) {
                return new Vector2(v0.v[0] / v1.v[0], v0.v[1] / v1.v[1]);
            }
            static scale(v0, scalar) {
                return new Vector2(v0.v[0] * scalar, v0.v[1] * scalar);
            }
            static negate(v0) {
                return new Vector2(-v0.v[0], -v0.v[1]);
            }
            static clone(v0) {
                return new Vector2(v0.v[0], v0.v[1]);
            }
            static create(x, y) {
                return new Vector2(x, y);
            }
        }
        exports.Vector2 = Vector2;
        Vector2.MAX_VALUE = new Vector2(f32.max, f32.max);
        Vector2.MIN_VALUE = new Vector2(f32.min, f32.min);
    });
    define("engine/math/vectorN", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class VectorN {
            constructor(array) {
                this.v = new Float32Array(array.length);
                this.v.set(array, 0);
            }
            toString() {
                const buffer = [];
                buffer.push('[');
                for (let i = 0; i < this.v.length; i++) {
                    buffer.push(i < (this.v.length - 1) ? `${this.v[i]}, ` : `${this.v[i]}`);
                }
                buffer.push(']');
                return buffer.join('');
            }
            kind() {
                return 'VectorN';
            }
            clone() {
                return VectorN.clone(this);
            }
            length() {
                return VectorN.getLength(this);
            }
            lengthSq() {
                return VectorN.getLengthSq(this);
            }
            normalize() {
                return VectorN.normalize(this);
            }
            dot(v0) {
                return VectorN.dot(this, v0);
            }
            add(v0) {
                return VectorN.add(this, v0);
            }
            sub(v0) {
                return VectorN.sub(this, v0);
            }
            mul(v0) {
                return VectorN.mul(this, v0);
            }
            div(v0) {
                return VectorN.div(this, v0);
            }
            scale(s0) {
                return VectorN.scale(this, s0);
            }
            negate() {
                return VectorN.negate(this);
            }
            static add(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] + v1.v[i];
                }
                return new VectorN(result);
            }
            static sub(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] - v1.v[i];
                }
                return new VectorN(result);
            }
            static mul(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] * v1.v[i];
                }
                return new VectorN(result);
            }
            static div(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] / v1.v[i];
                }
                return new VectorN(result);
            }
            static scale(v0, scalar) {
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] * scalar;
                }
                return new VectorN(result);
            }
            static distance(left, right) {
                if (left.v.length !== right.v.length)
                    throw Error('dimension mismatch.');
                let acc = 0;
                for (let i = 0; i < left.v.length; i++) {
                    const delta = right.v[i] - left.v[i];
                    acc += (delta * delta);
                }
                return Math.sqrt(acc);
            }
            static equals(left, right) {
                if (left.v.length !== right.v.length)
                    throw Error('dimension mismatch.');
                for (let i = 0; i < left.v.length; i++) {
                    if (left.v[i] !== right.v[i]) {
                        return false;
                    }
                }
                return true;
            }
            static clamp(v0, min, max) {
                const result = new Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i];
                    result[i] = (result[i] > max.v[i]) ? max.v[i] : result[i];
                    result[i] = (result[i] < min.v[i]) ? min.v[i] : result[i];
                }
                return new VectorN(result);
            }
            static lerp(v0, v1, amount) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] + ((v1.v[i] - v0.v[i]) * amount);
                }
                return new VectorN(result);
            }
            static getLength(v0) {
                let num = 0;
                for (let i = 0; i < v0.v.length; i++) {
                    num += (v0.v[i] * v0.v[i]);
                }
                return Math.sqrt(num);
            }
            static getLengthSq(v0) {
                let num = 0;
                for (let i = 0; i < v0.v.length; i++) {
                    num += (v0.v[i] * v0.v[i]);
                }
                return num;
            }
            static reflect(v0, n0) {
                if (v0.v.length !== n0.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                let n1 = 0;
                for (let i = 0; i < v0.v.length; i++)
                    n1 += (v0.v[i] * n0.v[i]);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = (v0.v[i] - ((2.0 * n1) * n0.v[i]));
                }
                return new VectorN(result);
            }
            static abs(v0) {
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = Math.abs(v0.v[i]);
                }
                return new VectorN(result);
            }
            static min(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] < v1.v[i] ? v0.v[i] : v1.v[i];
                }
                return new VectorN(result);
            }
            static max(v0, v1) {
                if (v0.v.length !== v1.v.length)
                    throw Error('dimension mismatch.');
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] > v1.v[i] ? v0.v[i] : v1.v[i];
                }
                return new VectorN(result);
            }
            static negate(v0) {
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = -v0.v[i];
                }
                return new VectorN(result);
            }
            static dot(v0, n0) {
                if (v0.v.length !== n0.v.length)
                    throw Error('dimension mismatch.');
                let num = 0;
                for (let i = 0; i < v0.v.length; i++) {
                    num += (v0.v[i] * n0.v[i]);
                }
                return num;
            }
            static normalize(v0) {
                const result = new Float32Array(v0.v.length);
                let n0 = 0;
                for (let i = 0; i < v0.v.length; i++)
                    n0 += (v0.v[i] * v0.v[i]);
                const n1 = 1.0 / Math.sqrt(n0);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i] * n1;
                }
                return new VectorN(result);
            }
            static clone(v0) {
                const result = new Float32Array(v0.v.length);
                for (let i = 0; i < v0.v.length; i++) {
                    result[i] = v0.v[i];
                }
                return new VectorN(result);
            }
            static create(...args) {
                return new VectorN(args);
            }
        }
        exports.VectorN = VectorN;
    });
    define("engine/math/index", ["require", "exports", "engine/math/angle", "engine/math/radian", "engine/math/matrix", "engine/math/plane", "engine/math/quaternion", "engine/math/single", "engine/math/vector2", "engine/math/vector3", "engine/math/vector4", "engine/math/vectorN", "engine/math/ray", "engine/math/triangle", "engine/math/box", "engine/math/sphere", "engine/math/frustum"], function (require, exports, angle_1, radian_1, matrix_2, plane_4, quaternion_1, single_2, vector2_1, vector3_8, vector4_1, vectorN_1, ray_3, triangle_1, box_1, sphere_1, frustum_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Angle = angle_1.Angle;
        exports.Radian = radian_1.Radian;
        exports.Matrix = matrix_2.Matrix;
        exports.Plane = plane_4.Plane;
        exports.Quaternion = quaternion_1.Quaternion;
        exports.Single = single_2.Single;
        exports.Vector2 = vector2_1.Vector2;
        exports.Vector3 = vector3_8.Vector3;
        exports.Vector4 = vector4_1.Vector4;
        exports.VectorN = vectorN_1.VectorN;
        exports.Ray = ray_3.Ray;
        exports.Triangle = triangle_1.Triangle;
        exports.Box = box_1.Box;
        exports.Sphere = sphere_1.Sphere;
        exports.Frustum = frustum_1.Frustum;
    });
    define("engine/raster/vertex", ["require", "exports", "engine/math/index"], function (require, exports, index_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Vertex {
            constructor(position, normal, uv) {
                this.position = position;
                this.normal = normal;
                this.uv = uv;
            }
            static correct(vertex, w, output) {
                output.position.v[0] = vertex.position.v[0] / w;
                output.position.v[1] = vertex.position.v[1] / w;
                output.position.v[2] = vertex.position.v[2] / w;
                output.position.v[3] = vertex.position.v[3] / w;
                output.normal.v[0] = vertex.normal.v[0] / w;
                output.normal.v[1] = vertex.normal.v[1] / w;
                output.normal.v[2] = vertex.normal.v[2] / w;
                output.uv.v[0] = vertex.uv.v[0] / w;
                output.uv.v[1] = vertex.uv.v[1] / w;
            }
            static interpolate(vertex_0, vertex_1, vertex_2, w_0, w_1, w_2, w, output) {
                output.position.v[0] = ((w_0 * vertex_0.position.v[0]) + (w_1 * vertex_1.position.v[0]) + (w_2 * vertex_2.position.v[0])) / w;
                output.position.v[1] = ((w_0 * vertex_0.position.v[1]) + (w_1 * vertex_1.position.v[1]) + (w_2 * vertex_2.position.v[1])) / w;
                output.position.v[2] = ((w_0 * vertex_0.position.v[2]) + (w_1 * vertex_1.position.v[2]) + (w_2 * vertex_2.position.v[2])) / w;
                output.position.v[3] = ((w_0 * vertex_0.position.v[3]) + (w_1 * vertex_1.position.v[3]) + (w_2 * vertex_2.position.v[3])) / w;
                output.normal.v[0] = ((w_0 * vertex_0.normal.v[0]) + (w_1 * vertex_1.normal.v[0]) + (w_2 * vertex_2.normal.v[0])) / w;
                output.normal.v[1] = ((w_0 * vertex_0.normal.v[1]) + (w_1 * vertex_1.normal.v[1]) + (w_2 * vertex_2.normal.v[1])) / w;
                output.normal.v[2] = ((w_0 * vertex_0.normal.v[2]) + (w_1 * vertex_1.normal.v[2]) + (w_2 * vertex_2.normal.v[2])) / w;
                output.uv.v[0] = ((w_0 * vertex_0.uv.v[0]) + (w_1 * vertex_1.uv.v[0]) + (w_2 * vertex_2.uv.v[0])) / w;
                output.uv.v[1] = ((w_0 * vertex_0.uv.v[1]) + (w_1 * vertex_1.uv.v[1]) + (w_2 * vertex_2.uv.v[1])) / w;
            }
            static create() {
                return new Vertex(index_1.Vector4.zero(), index_1.Vector3.zero(), index_1.Vector2.zero());
            }
        }
        exports.Vertex = Vertex;
    });
    define("engine/raster/shader", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
    });
    define("engine/raster/target", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
    });
    define("engine/raster/depth", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class DepthBuffer {
            constructor(width, height) {
                this.width = width;
                this.height = height;
                this.data = new Float32Array(this.width * this.height);
                this.clear();
            }
            clear() {
                for (let i = 0; i < this.data.length; i++) {
                    this.data[i] = Number.MIN_VALUE;
                }
            }
            set(x, y, value) {
                this.data[x + (y * this.width)] = value;
            }
            get(x, y) {
                return this.data[x + (y * this.width)];
            }
        }
        exports.DepthBuffer = DepthBuffer;
    });
    define("engine/raster/raster", ["require", "exports", "engine/math/index", "engine/raster/vertex"], function (require, exports, index_2, vertex_3) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Raster {
            static triangle(vertexProgram, fragmentProgram, depth, target, uniform, vertex_0, vertex_1, vertex_2) {
                const width = target.width;
                const height = target.height;
                const half_width = width * 0.5;
                const half_height = height * 0.5;
                vertexProgram.main(uniform, vertex_0, Raster.varying_0, Raster.position_0);
                vertexProgram.main(uniform, vertex_1, Raster.varying_1, Raster.position_1);
                vertexProgram.main(uniform, vertex_2, Raster.varying_2, Raster.position_2);
                if (Raster.position_0.v[2] < 0.0 || Raster.position_1.v[2] < 0.0 || Raster.position_2.v[2] < 0.0) {
                    return;
                }
                Raster.clippos_0.v[0] = ((Raster.position_0.v[0] / Raster.position_0.v[3]) * width) + half_width;
                Raster.clippos_0.v[1] = ((-Raster.position_0.v[1] / Raster.position_0.v[3]) * height) + half_height;
                Raster.clippos_1.v[0] = ((Raster.position_1.v[0] / Raster.position_1.v[3]) * width) + half_width;
                Raster.clippos_1.v[1] = ((-Raster.position_1.v[1] / Raster.position_1.v[3]) * height) + half_height;
                Raster.clippos_2.v[0] = ((Raster.position_2.v[0] / Raster.position_2.v[3]) * width) + half_width;
                Raster.clippos_2.v[1] = ((-Raster.position_2.v[1] / Raster.position_2.v[3]) * height) + half_height;
                vertex_3.Vertex.correct(Raster.varying_0, Raster.position_0.v[3], Raster.corrected_varying_0);
                vertex_3.Vertex.correct(Raster.varying_1, Raster.position_1.v[3], Raster.corrected_varying_1);
                vertex_3.Vertex.correct(Raster.varying_2, Raster.position_2.v[3], Raster.corrected_varying_2);
                if (Raster.edge(Raster.clippos_0, Raster.clippos_1, Raster.clippos_2) >= 0.0) {
                    Raster.draw_triangle(fragmentProgram, depth, target, uniform, Raster.corrected_varying_0, Raster.corrected_varying_1, Raster.corrected_varying_2, Raster.clippos_0, Raster.clippos_1, Raster.clippos_2, (1.0 / Raster.position_0.v[3]), (1.0 / Raster.position_1.v[3]), (1.0 / Raster.position_2.v[3]));
                }
            }
            static calculate_x_scan_range(y, ordered_0, ordered_1, ordered_2, ordered_3) {
                const gradient_0 = (ordered_0.v[1] !== ordered_1.v[1]) ? (y - ordered_0.v[1]) / (ordered_1.v[1] - ordered_0.v[1]) : 1.0;
                const gradient_1 = (ordered_2.v[1] !== ordered_3.v[1]) ? (y - ordered_2.v[1]) / (ordered_3.v[1] - ordered_2.v[1]) : 1.0;
                const min_x = ordered_0.v[0] + (ordered_1.v[0] - ordered_0.v[0]) * Raster.clamp(gradient_0, 0.0, 1.0);
                const max_x = ordered_2.v[0] + (ordered_3.v[0] - ordered_2.v[0]) * Raster.clamp(gradient_1, 0.0, 1.0);
                return [min_x | 0, max_x | 0];
            }
            static draw_triangle(fragment, depth, target, uniform, varying_0, varying_1, varying_2, clippos_0, clippos_1, clippos_2, corrected_z_0, corrected_z_1, corrected_z_2) {
                Raster.ordered_0.v[0] = clippos_0.v[0];
                Raster.ordered_0.v[1] = clippos_0.v[1];
                Raster.ordered_1.v[0] = clippos_1.v[0];
                Raster.ordered_1.v[1] = clippos_1.v[1];
                Raster.ordered_2.v[0] = clippos_2.v[0];
                Raster.ordered_2.v[1] = clippos_2.v[1];
                if (Raster.ordered_0.v[1] > Raster.ordered_1.v[1]) {
                    Raster.swap(Raster.ordered_0, Raster.ordered_1);
                }
                if (Raster.ordered_1.v[1] > Raster.ordered_2.v[1]) {
                    Raster.swap(Raster.ordered_1, Raster.ordered_2);
                }
                if (Raster.ordered_0.v[1] > Raster.ordered_1.v[1]) {
                    Raster.swap(Raster.ordered_0, Raster.ordered_1);
                }
                for (let y = (Raster.ordered_0.v[1] | 0); y <= (Raster.ordered_2.v[1] | 0); y++) {
                    if (y < Raster.ordered_1.v[1]) {
                        const [min_x, max_x] = Raster.calculate_x_scan_range(y, Raster.ordered_0, Raster.ordered_2, Raster.ordered_0, Raster.ordered_1);
                        Raster.draw_line(fragment, depth, target, uniform, clippos_0, clippos_1, clippos_2, varying_0, varying_1, varying_2, corrected_z_0, corrected_z_1, corrected_z_2, min_x, max_x, y);
                    }
                    else {
                        const [min_x, max_x] = Raster.calculate_x_scan_range(y, Raster.ordered_0, Raster.ordered_2, Raster.ordered_1, Raster.ordered_2);
                        Raster.draw_line(fragment, depth, target, uniform, clippos_0, clippos_1, clippos_2, varying_0, varying_1, varying_2, corrected_z_0, corrected_z_1, corrected_z_2, min_x, max_x, y);
                    }
                }
                for (let y = (Raster.ordered_0.v[1] | 0); y <= (Raster.ordered_2.v[1] | 0); y++) {
                    if (y < Raster.ordered_1.v[1]) {
                        const [min_x, max_x] = Raster.calculate_x_scan_range(y, Raster.ordered_0, Raster.ordered_1, Raster.ordered_0, Raster.ordered_2);
                        Raster.draw_line(fragment, depth, target, uniform, clippos_0, clippos_1, clippos_2, varying_0, varying_1, varying_2, corrected_z_0, corrected_z_1, corrected_z_2, min_x, max_x, y);
                    }
                    else {
                        const [min_x, max_x] = Raster.calculate_x_scan_range(y, Raster.ordered_1, Raster.ordered_2, Raster.ordered_0, Raster.ordered_2);
                        Raster.draw_line(fragment, depth, target, uniform, clippos_0, clippos_1, clippos_2, varying_0, varying_1, varying_2, corrected_z_0, corrected_z_1, corrected_z_2, min_x, max_x, y);
                    }
                }
            }
            static draw_line(fragment, depth, target, uniform, clippos_0, clippos_1, clippos_2, varying_0, varying_1, varying_2, corrected_z_0, corrected_z_1, corrected_z_2, min_x, max_x, y) {
                if (y < 0 || y >= target.height) {
                    return;
                }
                min_x = Math.max(min_x, 0);
                max_x = Math.min(max_x, target.width);
                const edge = Raster.edge(clippos_0, clippos_1, clippos_2);
                for (let x = min_x; x < max_x; x++) {
                    Raster.fragment_coord.v[0] = x;
                    Raster.fragment_coord.v[1] = y;
                    const weight_0 = Raster.edge(clippos_2, clippos_1, Raster.fragment_coord) / edge;
                    const weight_1 = Raster.edge(clippos_0, clippos_2, Raster.fragment_coord) / edge;
                    const weight_2 = Raster.edge(clippos_1, clippos_0, Raster.fragment_coord) / edge;
                    const calculated_depth = (weight_0 * corrected_z_0)
                        + (weight_1 * corrected_z_1)
                        + (weight_2 * corrected_z_2);
                    if (calculated_depth <= depth.get(x | 0, y | 0)) {
                        depth.set(x, y, calculated_depth);
                        vertex_3.Vertex.interpolate(varying_0, varying_1, varying_2, weight_0, weight_1, weight_2, calculated_depth, Raster.fragment_varying);
                        fragment.main(uniform, Raster.fragment_varying, Raster.fragment_color);
                        target.set(x, y, Raster.fragment_color);
                    }
                }
            }
            static edge(v0, v1, v2) {
                return (v2.v[0] - v0.v[0]) * (v1.v[1] - v0.v[1]) -
                    (v2.v[1] - v0.v[1]) * (v1.v[0] - v0.v[0]);
            }
            static clamp(f, min, max) {
                if (f < min)
                    return min;
                if (f > max)
                    return max;
                return f;
            }
            static swap(v0, v1) {
                const tx = v0.v[0];
                const ty = v0.v[1];
                v0.v[0] = v1.v[0];
                v0.v[1] = v1.v[1];
                v1.v[0] = tx;
                v1.v[1] = ty;
            }
        }
        exports.Raster = Raster;
        Raster.position_0 = index_2.Vector4.zero();
        Raster.position_1 = index_2.Vector4.zero();
        Raster.position_2 = index_2.Vector4.zero();
        Raster.clippos_0 = index_2.Vector2.zero();
        Raster.clippos_1 = index_2.Vector2.zero();
        Raster.clippos_2 = index_2.Vector2.zero();
        Raster.varying_0 = vertex_3.Vertex.create();
        Raster.varying_1 = vertex_3.Vertex.create();
        Raster.varying_2 = vertex_3.Vertex.create();
        Raster.corrected_varying_0 = vertex_3.Vertex.create();
        Raster.corrected_varying_1 = vertex_3.Vertex.create();
        Raster.corrected_varying_2 = vertex_3.Vertex.create();
        Raster.ordered_0 = index_2.Vector2.zero();
        Raster.ordered_1 = index_2.Vector2.zero();
        Raster.ordered_2 = index_2.Vector2.zero();
        Raster.fragment_varying = vertex_3.Vertex.create();
        Raster.fragment_coord = index_2.Vector2.zero();
        Raster.fragment_color = index_2.Vector4.zero();
    });
    define("engine/raster/index", ["require", "exports", "engine/raster/depth", "engine/raster/raster", "engine/raster/vertex"], function (require, exports, depth_1, raster_1, vertex_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.DepthBuffer = depth_1.DepthBuffer;
        exports.Raster = raster_1.Raster;
        exports.Vertex = vertex_1.Vertex;
    });
    define("engine/render/camera", ["require", "exports", "engine/math/index"], function (require, exports, index_3) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Camera {
            constructor(projection = index_3.Matrix.identity(), view = index_3.Matrix.identity()) {
                this.projection = projection;
                this.view = view;
            }
        }
        exports.Camera = Camera;
    });
    define("engine/render/geometry", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Geometry {
            constructor(vertices, indices) {
                this.vertices = vertices;
                this.indices = indices;
            }
        }
        exports.Geometry = Geometry;
    });
    define("engine/render/texture", ["require", "exports", "engine/math/index"], function (require, exports, index_4) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Texture {
            constructor(width, height) {
                this.width = width;
                this.height = height;
                this.data = new Float32Array(this.width * this.height * 4);
            }
            clear(color) {
                for (let offset = 0; offset < this.data.length; offset += 4) {
                    this.data.set(color.v, offset);
                }
            }
            sample(x, y) {
                const ax = (x < 0.0) ? Math.abs(x) : x;
                const ay = (y < 0.0) ? Math.abs(y) : y;
                const sx = ax * this.width;
                const sy = ay * this.height;
                const px = sx % this.width | 0;
                const py = sy % this.height | 0;
                const offset = this.offset(px, py);
                return index_4.Vector4.create(this.data[offset + 0], this.data[offset + 1], this.data[offset + 2], this.data[offset + 3]);
            }
            fast_sample(x, y, output) {
                const ax = (x < 0.0) ? Math.abs(x) : x;
                const ay = (y < 0.0) ? Math.abs(y) : y;
                const sx = ax * this.width;
                const sy = ay * this.height;
                const px = sx % this.width | 0;
                const py = sy % this.height | 0;
                const offset = this.offset(px, py);
                output.v[0] = this.data[offset + 0];
                output.v[1] = this.data[offset + 1];
                output.v[2] = this.data[offset + 2];
                output.v[3] = this.data[offset + 3];
            }
            fast_get(x, y, output) {
                const offset = this.offset(x, y);
                output.v[0] = this.data[offset + 0];
                output.v[1] = this.data[offset + 1];
                output.v[2] = this.data[offset + 2];
                output.v[3] = this.data[offset + 3];
            }
            get(x, y) {
                const offset = this.offset(x, y);
                return index_4.Vector4.create(this.data[offset + 0], this.data[offset + 1], this.data[offset + 2], this.data[offset + 3]);
            }
            set(x, y, color) {
                const offset = this.offset(x, y);
                this.data.set(color.v, offset);
            }
            offset(x, y) {
                const mx = (x % this.width);
                const my = (y % this.height);
                const rm = (mx + (my * this.width)) | 0;
                return rm * 4;
            }
        }
        exports.Texture = Texture;
    });
    define("engine/render/material", ["require", "exports", "engine/math/index"], function (require, exports, index_5) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Material {
            constructor(vertexProgram, fragmentProgram, uniforms) {
                this.vertexProgram = vertexProgram;
                this.fragmentProgram = fragmentProgram;
                this.uniforms = uniforms;
            }
        }
        exports.Material = Material;
        class TextureMaterialVertexShader {
            constructor() {
                this.matrix0 = index_5.Matrix.identity();
                this.matrix1 = index_5.Matrix.identity();
            }
            main(uniform, vertex, varying, output) {
                varying.position = vertex.position;
                varying.normal = vertex.normal;
                varying.uv = vertex.uv;
                index_5.Matrix.fast_mul(uniform.view, uniform.projection, this.matrix0);
                index_5.Matrix.fast_mul(uniform.matrix, this.matrix0, this.matrix1);
                index_5.Vector4.fast_transform(vertex.position, this.matrix1, output);
            }
        }
        exports.TextureMaterialVertexShader = TextureMaterialVertexShader;
        class TextureMaterialFragmentShader {
            main(uniform, varying, output) {
                uniform.texture.fast_sample(varying.uv.x, varying.uv.y, output);
            }
        }
        exports.TextureMaterialFragmentShader = TextureMaterialFragmentShader;
        class TextureMaterial extends Material {
            constructor(texture) {
                super(new TextureMaterialVertexShader(), new TextureMaterialFragmentShader(), { texture });
                this.texture = texture;
            }
        }
        exports.TextureMaterial = TextureMaterial;
    });
    define("engine/render/object", ["require", "exports", "engine/math/index"], function (require, exports, math_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Object3D {
            constructor() {
                this.matrix = math_1.Matrix.identity();
                this.objects = [];
            }
            kind() {
                return 'Object3D';
            }
        }
        exports.Object3D = Object3D;
    });
    define("engine/render/mesh", ["require", "exports", "engine/render/object"], function (require, exports, object_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Mesh extends object_1.Object3D {
            constructor(geometry, material) {
                super();
                this.geometry = geometry;
                this.material = material;
            }
            kind() {
                return "Mesh";
            }
        }
        exports.Mesh = Mesh;
    });
    define("engine/terminal/host", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        const SAFE_WIDTH = 80;
        const SAFE_HEIGHT = 24;
        class Host {
            static resize() {
                this.width = Host.get_width();
                this.height = Host.get_height();
            }
            static get_width() {
                return process.stdout.isTTY
                    ? process.stdout.columns
                    : SAFE_WIDTH;
            }
            static get_height() {
                if (process.stdout.isTTY) {
                    return (process.stdout.rows >= 3)
                        ? process.stdout.rows - 1
                        : 1;
                }
                else {
                    return SAFE_HEIGHT;
                }
            }
        }
        exports.Host = Host;
        Host._ = process.stdout.on('resize', () => Host.resize());
        Host.width = Host.get_width();
        Host.height = Host.get_height();
    });
    define("engine/terminal/terminal", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
    });
    define("engine/terminal/stream", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Stream {
            constructor(length) {
                this.buffer = Buffer.alloc(length);
                this.pointer = 0;
            }
            write(buffer) {
                this.buffer.set(buffer, this.pointer);
                this.pointer += buffer.length;
            }
            read() {
                const buffer = this.buffer.slice(0, this.pointer);
                this.pointer = 0;
                return buffer;
            }
        }
        exports.Stream = Stream;
    });
    define("engine/terminal/ascii", ["require", "exports", "engine/math/index", "engine/terminal/stream", "engine/terminal/host"], function (require, exports, index_6, stream_1, host_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class AsciiTerminal {
            constructor() {
                this.size = { width: 0, height: 0 };
                this.setup_buffers();
            }
            get width() {
                return this.size.width;
            }
            get height() {
                return this.size.height;
            }
            async present(texture) {
                this.assert_buffers();
                const color = index_6.Vector4.zero();
                for (let y = 0; y < this.size.height; y++) {
                    for (let x = 0; x < this.size.width; x++) {
                        texture.fast_get(x, y, color);
                        this.write_pixel(color);
                    }
                    this.stream.write(this.ansi_newline);
                }
                this.stream.write(this.ansi_reset);
                const buffer = this.stream.read();
                process.stdout.write(buffer);
            }
            clamp(x) {
                if (x <= 0.0) {
                    return 0.00;
                }
                if (x >= 1.0) {
                    return 0.99;
                }
                return x;
            }
            compute_gradient_index(color) {
                const average = (color.v[0] + color.v[1] + color.v[2]) / 3;
                const shade = this.clamp(average);
                return (shade * (this.ansi_gradient.length - 1)) | 0;
            }
            write_pixel(color) {
                const index = this.compute_gradient_index(color);
                const character = this.ansi_gradient[index];
                this.stream.write(character);
            }
            setup_buffers() {
                this.size = { width: host_1.Host.width, height: host_1.Host.height };
                this.stream = new stream_1.Stream(2000000);
                this.ansi_gradient = ' .:;+=xX$#&'.split('').map(n => Buffer.from(n));
                this.ansi_newline = Buffer.from('\n');
                this.ansi_reset = Buffer.concat([
                    Buffer.from(`\x1b[${this.size.width}D`),
                    Buffer.from(`\x1b[${this.size.height}A`)
                ]);
            }
            assert_buffers() {
                if (host_1.Host.width !== this.size.width || host_1.Host.height !== this.size.height) {
                    this.size = { width: host_1.Host.width, height: host_1.Host.height };
                    this.ansi_reset = Buffer.concat([
                        Buffer.from(`\x1b[${this.size.width}D`),
                        Buffer.from(`\x1b[${this.size.height}A`)
                    ]);
                }
            }
        }
        exports.AsciiTerminal = AsciiTerminal;
    });
    define("engine/terminal/color", ["require", "exports", "engine/math/index", "engine/terminal/stream", "engine/terminal/host"], function (require, exports, index_7, stream_2, host_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class ColorTerminal {
            constructor() {
                this.size = { width: 0, height: 0 };
                this.setup_buffers();
            }
            get width() {
                return this.size.width;
            }
            get height() {
                return this.size.height;
            }
            async present(texture) {
                this.assert_buffers();
                const color = index_7.Vector4.zero();
                for (let y = 0; y < this.size.height; y++) {
                    for (let x = 0; x < this.size.width; x++) {
                        texture.fast_get(x, y, color);
                        this.write_pixel(color);
                    }
                    this.stream.write(this.ansi_newline);
                }
                this.stream.write(this.ansi_reset);
                const buffer = this.stream.read();
                process.stdout.write(buffer);
            }
            write_pixel(color) {
                const r = Math.floor(color.v[0] * 255);
                const g = Math.floor(color.v[1] * 255);
                const b = Math.floor(color.v[2] * 255);
                this.stream.write(this.ansi_begin);
                this.stream.write(this.ansi_numerics[r]);
                this.stream.write(this.ansi_delimiter);
                this.stream.write(this.ansi_numerics[g]);
                this.stream.write(this.ansi_delimiter);
                this.stream.write(this.ansi_numerics[b]);
                this.stream.write(this.ansi_end);
            }
            setup_buffers() {
                this.size = { width: host_2.Host.width, height: host_2.Host.height };
                this.stream = new stream_2.Stream(8000000);
                this.ansi_begin = Buffer.from(`\x1b[48;2;`);
                this.ansi_numerics = Array.from({ length: 256 }).map((_, i) => Buffer.from(i.toString()));
                this.ansi_delimiter = Buffer.from(';');
                this.ansi_end = Buffer.from(`m `);
                this.ansi_newline = Buffer.from('\x1b[48;2;0;0;0m\n');
                this.ansi_reset = Buffer.concat([
                    Buffer.from(`\x1b[${this.size.width}D`),
                    Buffer.from(`\x1b[${this.size.height}A`)
                ]);
            }
            assert_buffers() {
                if (host_2.Host.width !== this.size.width || host_2.Host.height !== this.size.height) {
                    this.size = { width: host_2.Host.width, height: host_2.Host.height };
                    this.ansi_reset = Buffer.concat([
                        Buffer.from(`\x1b[${this.size.width}D`),
                        Buffer.from(`\x1b[${this.size.height}A`)
                    ]);
                }
            }
        }
        exports.ColorTerminal = ColorTerminal;
    });
    define("engine/terminal/index", ["require", "exports", "engine/terminal/host", "engine/terminal/ascii", "engine/terminal/color"], function (require, exports, host_3, ascii_1, color_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Host = host_3.Host;
        exports.AsciiTerminal = ascii_1.AsciiTerminal;
        exports.ColorTerminal = color_1.ColorTerminal;
    });
    define("engine/render/scene", ["require", "exports", "engine/render/object"], function (require, exports, object_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Scene extends object_2.Object3D {
            constructor() {
                super();
            }
            kind() {
                return "Scene";
            }
            add(object) {
                this.objects.push(object);
            }
        }
        exports.Scene = Scene;
    });
    define("engine/render/renderer", ["require", "exports", "engine/math/index", "engine/raster/index", "engine/raster/index", "engine/render/texture"], function (require, exports, index_8, index_9, index_10, texture_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Renderer {
            constructor(terminal) {
                this.terminal = terminal;
                this.size = { width: 0, height: 0 };
                this.setup_buffers();
            }
            render_object(camera, object3D, transform) {
                for (const object of object3D.objects) {
                    const matrix = index_8.Matrix.mul(transform, object.matrix);
                    switch (object.kind()) {
                        case "Object3D":
                            this.render_object(camera, object, matrix);
                            break;
                        case "Scene":
                            this.render_scene(camera, object, matrix);
                            break;
                        case "Mesh":
                            this.render_mesh(camera, object, matrix);
                            break;
                    }
                }
            }
            render_scene(camera, scene, transform) {
                for (const object of scene.objects) {
                    const matrix = index_8.Matrix.mul(transform, object.matrix);
                    switch (object.kind()) {
                        case "Object3D":
                            this.render_object(camera, object, matrix);
                            break;
                        case "Scene":
                            this.render_scene(camera, object, matrix);
                            break;
                        case "Mesh":
                            this.render_mesh(camera, object, matrix);
                            break;
                    }
                }
            }
            render_mesh(camera, mesh, transform) {
                const geometry = mesh.geometry;
                const material = mesh.material;
                const matrix = index_8.Matrix.mul(transform, mesh.matrix);
                const projection = camera.projection;
                const view = camera.view;
                const uniform = { projection, view, matrix, ...material.uniforms };
                for (let i = 0; i < geometry.indices.length; i += 3) {
                    const vertex_0 = geometry.vertices[geometry.indices[i + 0]];
                    const vertex_1 = geometry.vertices[geometry.indices[i + 1]];
                    const vertex_2 = geometry.vertices[geometry.indices[i + 2]];
                    index_10.Raster.triangle(material.vertexProgram, material.fragmentProgram, this.depth_buffer, this.color_buffer, uniform, vertex_0, vertex_1, vertex_2);
                }
            }
            clear(color) {
                this.assert_buffers();
                this.color_buffer.clear(color);
                this.depth_buffer.clear();
            }
            render(camera, scene) {
                this.assert_buffers();
                this.render_scene(camera, scene, scene.matrix);
                this.terminal.present(this.color_buffer);
            }
            setup_buffers() {
                this.size = { width: this.terminal.width, height: this.terminal.height };
                this.depth_buffer = new index_9.DepthBuffer(this.size.width, this.size.height);
                this.color_buffer = new texture_1.Texture(this.size.width, this.size.height);
            }
            assert_buffers() {
                if (this.terminal.width !== this.size.width ||
                    this.terminal.height !== this.size.height) {
                    this.setup_buffers();
                }
            }
        }
        exports.Renderer = Renderer;
    });
    define("engine/render/index", ["require", "exports", "engine/render/camera", "engine/render/geometry", "engine/render/material", "engine/render/mesh", "engine/render/object", "engine/render/renderer", "engine/render/scene", "engine/render/texture"], function (require, exports, camera_1, geometry_1, material_1, mesh_1, object_3, renderer_1, scene_1, texture_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Camera = camera_1.Camera;
        exports.Geometry = geometry_1.Geometry;
        exports.Material = material_1.Material;
        exports.TextureMaterial = material_1.TextureMaterial;
        exports.Mesh = mesh_1.Mesh;
        exports.Object3D = object_3.Object3D;
        exports.Renderer = renderer_1.Renderer;
        exports.Scene = scene_1.Scene;
        exports.Texture = texture_2.Texture;
    });
    define("engine/timing/stopwatch", ["require", "exports"], function (require, exports) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        class Stopwatch {
            constructor(samples = 32) {
                this.samples = Array.from({ length: samples }, () => 0);
                this.current = Date.now();
                this.index = 0;
            }
            start() {
                this.current = Date.now();
            }
            stop() {
                const delta = Date.now() - this.current;
                this.samples[this.index] = delta === 0 ? 1000 : (1000 / delta);
                this.index = (this.index + 1) % this.samples.length;
            }
            get() {
                const total = this.samples.reduce((acc, c) => acc + c, 0);
                return (total / this.samples.length) | 0;
            }
        }
        exports.Stopwatch = Stopwatch;
    });
    define("engine/timing/index", ["require", "exports", "engine/timing/stopwatch"], function (require, exports, stopwatch_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Stopwatch = stopwatch_1.Stopwatch;
    });
    define("engine/index", ["require", "exports", "engine/animation/index", "engine/math/index", "engine/raster/index", "engine/render/index", "engine/terminal/index", "engine/timing/index"], function (require, exports, index_11, index_12, index_13, index_14, index_15, index_16) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        exports.Animation = index_11.Animation;
        exports.Angle = index_12.Angle;
        exports.Box = index_12.Box;
        exports.Frustum = index_12.Frustum;
        exports.Matrix = index_12.Matrix;
        exports.Plane = index_12.Plane;
        exports.Quaternion = index_12.Quaternion;
        exports.Radian = index_12.Radian;
        exports.Ray = index_12.Ray;
        exports.Single = index_12.Single;
        exports.Sphere = index_12.Sphere;
        exports.Triangle = index_12.Triangle;
        exports.Vector2 = index_12.Vector2;
        exports.Vector3 = index_12.Vector3;
        exports.Vector4 = index_12.Vector4;
        exports.VectorN = index_12.VectorN;
        exports.DepthBuffer = index_13.DepthBuffer;
        exports.Raster = index_13.Raster;
        exports.Vertex = index_13.Vertex;
        exports.Camera = index_14.Camera;
        exports.Geometry = index_14.Geometry;
        exports.Material = index_14.Material;
        exports.Mesh = index_14.Mesh;
        exports.Object3D = index_14.Object3D;
        exports.Renderer = index_14.Renderer;
        exports.Scene = index_14.Scene;
        exports.Texture = index_14.Texture;
        exports.TextureMaterial = index_14.TextureMaterial;
        exports.AsciiTerminal = index_15.AsciiTerminal;
        exports.ColorTerminal = index_15.ColorTerminal;
        exports.Stopwatch = index_16.Stopwatch;
    });
    define("loader", ["require", "exports", "engine/index", "engine/index", "jpeg-js", "fs", "path"], function (require, exports, engine_1, engine_2, jpeg_js_1, fs_1, path_1) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        function load_texture(filename) {
            const image = jpeg_js_1.decode(fs_1.readFileSync(filename));
            const buffer = new engine_1.Texture(image.width, image.height);
            for (let y = 0; y < image.height; y++) {
                for (let x = 0; x < image.width; x++) {
                    const offset = (x + (y * image.width)) * 4;
                    const r = image.data[offset + 0] / 255;
                    const g = image.data[offset + 1] / 255;
                    const b = image.data[offset + 2] / 255;
                    const a = image.data[offset + 3] / 255;
                    const c = engine_2.Vector4.create(r, g, b, a);
                    buffer.set(x, (image.height - 1) - y, c);
                }
            }
            return buffer;
        }
        exports.load_texture = load_texture;
        function load_geometry(filepath) {
            const acc_v = [];
            const acc_vn = [];
            const acc_vt = [];
            const positions = [];
            const normals = [];
            const uvs = [];
            const indices = [];
            const content = fs_1.readFileSync(filepath, 'utf8');
            for (const line of content.split('\n')) {
                const parts = line.split(" ").map(p => p.trim());
                if (parts.length > 0) {
                    switch (parts[0]) {
                        case "v": {
                            const x = parseFloat(parts[1]);
                            const y = parseFloat(parts[2]);
                            const z = parseFloat(parts[3]);
                            acc_v.push(engine_2.Vector4.create(x, y, z, 1.0));
                            break;
                        }
                        case "vn": {
                            const x = parseFloat(parts[1]);
                            const y = parseFloat(parts[2]);
                            const z = parseFloat(parts[3]);
                            acc_vn.push(engine_2.Vector3.create(x, y, z));
                            break;
                        }
                        case "vt": {
                            const x = parseFloat(parts[1]);
                            const y = parseFloat(parts[2]);
                            acc_vt.push(engine_2.Vector2.create(x, y));
                            break;
                        }
                        case "f": {
                            for (let i = 1; i <= 3; i++) {
                                const face = parts[i].split("/");
                                const i_v = parseInt(face[0]) - 1;
                                const i_vt = parseInt(face[1]) - 1;
                                const i_vn = parseInt(face[2]) - 1;
                                if (i_v > acc_v.length - 1) {
                                    throw Error(`Invalid position index for face: ${parts[0]}`);
                                }
                                if (i_vt > acc_vt.length - 1) {
                                    throw Error(`Invalid texcoord index for face: ${parts[0]}`);
                                }
                                if (i_vn > acc_vn.length - 1) {
                                    throw Error(`Invalid normal index for face: ${parts[0]}`);
                                }
                                positions.push(acc_v[i_v]);
                                normals.push(acc_vn[i_vn]);
                                uvs.push(acc_vt[i_vt]);
                                indices.push(indices.length);
                            }
                        }
                        default: break;
                    }
                }
            }
            const vertices = [];
            for (let i = 0; i < positions.length; i++) {
                vertices.push(new engine_1.Vertex(positions[i], normals[i], uvs[i]));
            }
            return new engine_1.Geometry(vertices, indices);
        }
        exports.load_geometry = load_geometry;
        function load_animation(directory) {
            const framedata = JSON.parse(fs_1.readFileSync(path_1.join(directory, 'animation.json'), 'utf-8'));
            const animation = new engine_1.Animation();
            for (const frame of framedata) {
                animation.add('camera', {
                    timestamp: frame.frame_index,
                    properties: {
                        position: [
                            frame.properties.position_x,
                            frame.properties.position_z,
                            -frame.properties.position_y
                        ],
                        target: [
                            frame.properties.target_x,
                            frame.properties.target_z,
                            -frame.properties.target_y
                        ],
                        up: [0, 1, 0]
                    }
                });
            }
            return animation;
        }
        exports.load_animation = load_animation;
        function load_scene(directory) {
            const contents = fs_1.readdirSync(directory);
            const models = contents.filter(file => path_1.extname(file) === '.obj');
            const scene = new engine_1.Scene();
            for (const model_name of models) {
                const texture_name = model_name.slice(0, model_name.length - 4) + '.jpg';
                if (contents.includes(texture_name)) {
                    console.log(' loading ... ', model_name);
                    const geometry = load_geometry(path_1.join(directory, model_name));
                    console.log(' loading ... ', texture_name);
                    const material = new engine_1.TextureMaterial(load_texture(path_1.join(directory, texture_name)));
                    const mesh = new engine_1.Mesh(geometry, material);
                    scene.add(mesh);
                }
            }
            console.log('');
            return scene;
        }
        exports.load_scene = load_scene;
    });
    define("index", ["require", "exports", "engine/index", "engine/index", "engine/index", "engine/index", "loader", "path"], function (require, exports, index_17, index_18, index_19, index_20, loader_1, path_2) {
        "use strict";
        Object.defineProperty(exports, "__esModule", { value: true });
        console.log(`
                                             .___                     
    ________ ___________  ____             __| _/____   _____   ____  
    \\___   // __ \\_  __ \\/  _ \\   ______  / __ |/ __ \\ /     \\ /  _ \\ 
     /    /\\  ___/|  | \\(  <_> ) /_____/ / /_/ \\  ___/|  Y Y  (  <_> )
    /_____ \\\\___  >__|   \\____/          \\____ |\\___  >__|_|  /\\____/ 
          \\/    \\/                            \\/    \\/      \\/        
                                         
          reduce your terminal font size for better resolution                                                   
    `);
        const animation = loader_1.load_animation(path_2.join(__dirname, './scene'));
        const scene = loader_1.load_scene(path_2.join(__dirname, './scene'));
        const terminal = process.argv.includes('--color') ? new index_20.ColorTerminal() : new index_20.AsciiTerminal();
        const stopwatch = new index_17.Stopwatch();
        const renderer = new index_18.Renderer(terminal);
        const camera = new index_18.Camera();
        async function loop() {
            process.title = `zero-demo: ${terminal.width} x ${terminal.height} @ ${stopwatch.get()} fps - Sinclair: November 2019`;
            const aspect = terminal.width / (terminal.height * 2.0);
            const state = animation.state('camera', (Date.now() / 24) % animation.length('camera'));
            camera.projection = index_19.Matrix.perspectiveFov(index_18.Radian.fromAngle(90), aspect, 0.1, 1000);
            camera.view = index_19.Matrix.lookAt(index_19.Vector3.create(state.position[0], state.position[1], state.position[2]), index_19.Vector3.create(state.target[0], state.target[1], state.target[2]), index_19.Vector3.create(state.up[0], state.up[1], state.up[2]));
            stopwatch.start();
            renderer.clear(index_19.Vector4.create(1, 1, 1, 0));
            renderer.render(camera, scene);
            stopwatch.stop();
            setImmediate(() => loop());
        }
        loop();
    });
    
    'marker:resolver';

    function get_define(name) {
        if (defines[name]) {
            return defines[name];
        }
        else if (defines[name + '/index']) {
            return defines[name + '/index'];
        }
        else {
            const dependencies = ['exports'];
            const factory = (exports) => {
                try {
                    Object.defineProperty(exports, "__cjsModule", { value: true });
                    Object.defineProperty(exports, "default", { value: require(name) });
                }
                catch {
                    throw Error(['module "', name, '" not found.'].join(''));
                }
            };
            return { dependencies, factory };
        }
    }
    const instances = {};
    function resolve(name) {
        if (instances[name]) {
            return instances[name];
        }
        if (name === 'exports') {
            return {};
        }
        const define = get_define(name);
        instances[name] = {};
        const dependencies = define.dependencies.map(name => resolve(name));
        define.factory(...dependencies);
        const exports = dependencies[define.dependencies.indexOf('exports')];
        instances[name] = (exports['__cjsModule']) ? exports.default : exports;
        return instances[name];
    }
    if (entry[0] !== null) {
        return resolve(entry[0]);
    }
})();